import os
import sys
import contextlib
import io


@contextlib.contextmanager
def suppress_omp_nested_warning():
    try:
        # This works only outside of Jupyter
        original_stdout_fd = sys.stdout.fileno()
        original_stderr_fd = sys.stderr.fileno()

        with open(os.devnull, 'w') as devnull:
            devnull_fd = devnull.fileno()

            # Save original fds
            saved_stdout = os.dup(original_stdout_fd)
            saved_stderr = os.dup(original_stderr_fd)

            try:
                os.dup2(devnull_fd, original_stdout_fd)
                os.dup2(devnull_fd, original_stderr_fd)
                yield
            finally:
                os.dup2(saved_stdout, original_stdout_fd)
                os.dup2(saved_stderr, original_stderr_fd)
                os.close(saved_stdout)
                os.close(saved_stderr)

    except (io.UnsupportedOperation, AttributeError):
        # Fallback for Jupyter: suppress Python-level output only
        with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
            yield



import paicos as pa
import numpy as np
from scipy.spatial import KDTree
from paicos.trees.bvh_cpu import BinaryTree
from paicos.trees.bvh_cpu import get_morton_keys64_old, get_morton_keys64
from paicos.trees.bvh_cpu import propagate_bounds_upwards_old, propagate_bounds_upwards
from paicos.trees.bvh_cpu import generateHierarchy_old, generateHierarchy
from paicos.trees.bvh_cpu import set_leaf_bounding_volumes_old, set_leaf_bounding_volumes

pa.use_units(False)

snap = pa.Snapshot(pa.data_dir, 247, basename='reduced_snap',
                   load_catalog=False)
center = np.array([398968.4, 211682.6, 629969.9])
widths = np.array([2000, 2000, 2000])

# Select a cube
index = pa.util.get_index_of_cubic_region_plus_thin_layer(
    snap['0_Coordinates'], center, widths,
    snap['0_Diameters'], snap.box)

snap = snap.select(index, parttype=0)

pos = snap['0_Coordinates']
sizes = snap['0_Diameters']

L = 21

positions = pos
sizes = 1.2 * sizes

_pos = np.array(positions)

# Convert _position to units from [0, 1) * 2**L
off_sets = np.min(_pos, axis=0)
_pos -= off_sets[None, :]

max_pos = np.max(_pos)
conversion_factor = (2**L - 1) / max_pos
_pos *= conversion_factor

# if use_units take the value...
# pos = pos.value
#

# Calculate uint64 and Morton keys
_pos_uint = _pos.astype(np.uint64)

morton_keys_old = get_morton_keys64_old(_pos_uint)
morton_keys = get_morton_keys64(_pos_uint)

np.testing.assert_array_equal(morton_keys, morton_keys_old)
# print(morton_keys)

# Store index for going back to original ids.
sort_index = np.argsort(morton_keys)

# Do the sorting
morton_keys = morton_keys[sort_index]
_pos = _pos[sort_index, :]
_pos_uint = _pos_uint[sort_index, :]

# # Some tree properties
num_leafs = morton_keys.shape[0]
num_internal_nodes = num_leafs - 1
num_leafs_and_nodes = 2 * num_leafs - 1

# Allocate arrays
children = -1 * np.ones((num_internal_nodes, 2), dtype=int)
parents = -1 * np.ones(num_leafs_and_nodes, dtype=int)

children_old = -1 * np.ones((num_internal_nodes, 2), dtype=int)
parents_old = -1 * np.ones(num_leafs_and_nodes, dtype=int)

# # This sets the parent and children properties
generateHierarchy(morton_keys, children, parents)
generateHierarchy_old(morton_keys, children_old, parents_old)

np.testing.assert_array_equal(parents, parents_old)
np.testing.assert_array_equal(children, children_old)

# Set the boundaries for the leafs
bounds_old = np.zeros((num_leafs_and_nodes, 3, 2), dtype=np.uint64)
bounds = np.zeros((num_leafs_and_nodes, 3, 2), dtype=np.uint64)
bounds_cython = np.zeros((num_leafs_and_nodes, 3, 2), dtype=np.uint64)

set_leaf_bounding_volumes_old(bounds_old, _pos,
                              sizes[sort_index],
                              conversion_factor)
set_leaf_bounding_volumes(bounds, _pos,
                          sizes[sort_index],
                          conversion_factor, L)

np.testing.assert_array_equal(bounds, bounds_old)

propagate_bounds_upwards_old(bounds_old, parents, children)
propagate_bounds_upwards(bounds, parents, children)

np.testing.assert_array_equal(bounds, bounds_old)



"""
Defines small_a and small_h, conversion to Gauss and eV,
and the PaicosQuantity and PaicosTimeSeries classes.
"""
from fractions import Fraction
import numpy as np
import astropy.units as u
from astropy.units import Quantity
from astropy.units import UnitConversionError

_ns = globals()

small_a = u.def_unit(
    ["small_a"],
    prefixes=False,
    namespace=_ns,
    doc="Cosmological scale factor.",
    format={"latex": "a", "unicode": r"a"},
)
u.def_physical_type(small_a, "scale factor")


small_h = u.def_unit(
    ["small_h"],
    namespace=_ns,
    prefixes=False,
    doc='Reduced/"dimensionless" Hubble constant',
    format={"latex": r"h", "unicode": r"h"},
)

u.add_enabled_units(small_a)
u.add_enabled_units(small_h)

# Allows conversion between K and eV
u.add_enabled_equivalencies(u.temperature_energy())
# Allows conversion to Gauss (potential issues?)
# https://github.com/astropy/astropy/issues/7396

# The equivalencies sometimes frustratingly fail.
# TODO: Override the .to method ignore the Gauss component of the unit...
gauss_B = (u.g / u.cm)**(0.5) / u.s
equiv_B = [(u.G, gauss_B, lambda x: x, lambda x: x)]
Bscaling = small_a**(-2) * small_h
equiv_B_comoving = [(u.G * Bscaling, gauss_B * Bscaling, lambda x: x, lambda x: x)]
Bscaling = small_a**(-2)
equiv_B_no_small_h = [(u.G * Bscaling, gauss_B * Bscaling, lambda x: x, lambda x: x)]
u.add_enabled_equivalencies(equiv_B)
u.add_enabled_equivalencies(equiv_B_comoving)
u.add_enabled_equivalencies(equiv_B_no_small_h)


def get_unit_dictionaries(unit):
    """
    Returns dictionaries with information about the units of the
    quantity.

    :meta private:
    """
    codic = {}
    dic = {}
    for base, power in zip(unit.bases, unit.powers):
        if base in (small_a, small_h):
            codic[base] = power
        else:
            dic[base] = power
    for key in [small_h, small_a]:
        if key not in codic:
            codic[key] = 0

    return codic, dic


def construct_unit_from_dic(dic):
    """
    Construct unit from a dictionary with the format returned
    from get_unit_dictionaries

    :meta private:
    """
    return np.prod([unit**dic[unit] for unit in dic])


def separate_units(unit):
    """
    Separates the standard physical units (u_unit) from the units involving
    a and h (pu_unit). That is::

        u_unit, pu_unit = separate_units(unit)

    where pu_unit contains small_a and small_h and u_unit contains
    everything else such that::

        unit = u_unit * pu_unit
    """
    codic, dic = get_unit_dictionaries(unit)
    u_unit = construct_unit_from_dic(dic)
    pu_unit = construct_unit_from_dic(codic)
    return u_unit, pu_unit


def get_new_unit(unit, remove_list=[]):
    """
    Return new unit where base units in the remove list have been removed.

    :meta private:
    """
    unit_list = []
    for base, power in zip(unit.bases, unit.powers):
        if base not in remove_list:
            unit_list.append(base**power)

    return np.prod(unit_list)


class PaicosQuantity(Quantity):

    """
    PaicosQuantity is a subclass of the astropy Quantity class which
    represents a number with some associated unit.

    This subclass in addition includes a and h factors used in the definition
    of comoving variables.

    Parameters
    ----------

    value: the numeric values of your data (similar to astropy Quantity)

    a: the cosmological scale factor of your data

    h: the reduced Hubble parameter, e.g. h = 0.7

    unit: a string, e.g. ``g/cm^3 small_a^-3 small_h^2`` or an astropy Unit
          The latter can be defined like this::

            from paicos import units as pu
            from astropy import units as u
            unit = u.g*u.cm**(-3)*small_a**(-3)*small_h**(2)

          The naming of small_a and small_h is to avoid conflict with the already
          existing 'annum' (i.e. a year) and 'h' (hour) units.


    Methods/key properties
    ----------------------

        no_small_h
                    Returns a new comoving quantity where the h-factors
                    have been removed and the numeric value adjusted
                    accordingly.

        to_physical
                      Returns a new  object where both a and h factors have been
                      removed, i.e. we have switched from comoving values to
                      the physical value.

        label
                Returns a Latex label for use in plots.

    Examples
    --------

    Here is an example::

        units = 'g cm^-3 small_a^-3 small_h^2'
        A = PaicosQuantity(2, units, h=0.7, a=1/128)

        # Create a new comoving quantity where the h-factors have been removed
        B = A.no_small_h

        # Create a new quantity where both a and h factor have been removed,
        # i.e. we have switched from a comoving quantity to the physical value

        C = A.to_physical

    """

    # pylint: disable=too-many-arguments

    def __new__(cls, value, unit=None, dtype=None, copy=False, order=None,
                subok=False, ndmin=0, h=None, a=None, comoving_sim=None):
        """
        Here we initialize the Paicos Quantity. The three additional
        input arguments (compared to a standard astropy quantity) are

        h
        a
        comoving_sim

        :meta private:
        """

        assert h is not None, 'Paicos quantity is missing a value for h'
        assert a is not None, 'Paicos quantity is missing a value for a'
        assert comoving_sim is not None, 'is this from a comoving_sim?'

        if hasattr(value, 'unit'):
            unit = value.unit
            value = value.value

        obj = super().__new__(cls, value, unit=unit, dtype=dtype, copy=copy,
                              order=order, subok=subok, ndmin=ndmin)

        obj._h = h
        obj._a = a
        obj._comoving_sim = comoving_sim

        return obj

    def __array_finalize__(self, obj):
        """
        Heavily inspired by the astropy Quantity version.

        :meta private:
        """
        super_array_finalize = super().__array_finalize__
        if super_array_finalize is not None:
            super_array_finalize(obj)

        # If we're a new object or viewing an ndarray, nothing has to be done.
        if obj is None or obj.__class__ is np.ndarray:
            return

        # Set Paicos specific parameters
        self._h = getattr(obj, '_h', None)
        self._a = getattr(obj, '_a', None)
        self._comoving_sim = getattr(obj, '_comoving_sim', None)

    @property
    def a(self):
        """
        The scale factor.
        """
        if self.comoving_sim:
            return self._a
        raise RuntimeError('Non-comoving object has no scale factor')

    @property
    def h(self):
        """
        The reduced Hubble parameter
        """
        return self._h

    @property
    def comoving_sim(self):
        """
        Whether the simulation had ComovingIntegrationOn
        """
        return self._comoving_sim

    @property
    def z(self):
        """
        The redshift.
        """
        if self.comoving_sim:
            return 1. / self._a - 1.
        raise RuntimeError('Non-comoving object has no redshift')

    def lookback_time(self, reader_object):
        """
        The lookback time.

        Requires a reader object as input (e.g. a snap,
        cat or other instance of Snapshot, Catalog or PaicosReader)
        """
        if self.comoving_sim:
            return reader_object.get_lookback_time(self.z)
        msg = 'lookback_time not defined for non-comoving sim'
        raise RuntimeError(msg)

    def age(self, reader_object):
        """
        The age of the universe in the simulation.

        Requires a reader object as input (e.g. a snap,
        cat or other instance of Snapshot, Catalog or PaicosReader)
        """
        if self.comoving_sim:
            return reader_object.get_age(self.z)
        msg = 'age not defined for non-comoving sim'
        raise RuntimeError(msg)

    @property
    def time(self):
        """
        The time elapsed since the beginning of the simulation.

        Only defined for non-comoving simulations.
        """
        if self.comoving_sim:
            msg = 'time not defined for comoving sim'
            raise RuntimeError(msg)

        if hasattr(self._a, 'unit'):
            return self._a
        else:
            time = self._a * u.Unit('arepo_time')
            if self._h != 1.0:
                time = PaicosQuantity(self._a, 'arepo_time/small_h', a=self._a, h=self.h,
                                      comoving_sim=self.comoving_sim, copy=True)
            else:
                time = self._a * u.Unit('arepo_time')
            return time

    @property
    def unit_quantity(self):
        """
        Returns a new PaicosQuantity with the same units as the current
        PaicosQuantity and a numeric value of 1.
        """

        return PaicosQuantity(1., self.unit, a=self._a, h=self.h, copy=True,
                              comoving_sim=self.comoving_sim)

    @property
    def copy(self):
        """
        Returns a copy of the PaicosQuantity.
        """
        return PaicosQuantity(np.array(self.value), self.unit, a=self._a, h=self.h,
                              comoving_sim=self.comoving_sim, copy=True)

    @property
    def uq(self):
        """
        A short hand for the 'unit_quantity' method.
        """
        return self.unit_quantity

    @property
    def hdf5_attrs(self):
        """
        Give the units as a dictionary for hdf5 data set attributes

        :meta private:
        """
        return {'unit': self.unit.to_string()}

    @property
    def no_small_h(self):
        """
        Remove scaling with h, returning a quantity with adjusted values.
        """
        codic, _ = get_unit_dictionaries(self.unit)
        factor = self.h**codic[small_h]

        value = self.view(np.ndarray)
        new_unit = get_new_unit(self.unit, [small_h])
        return self._new_view(value * factor, new_unit)

    @property
    def cgs(self):
        """
        Returns a copy of the current `PaicosQuantity` instance with CGS units.
        The value of the resulting object will be scaled.
        """
        u_unit, pu_unit = separate_units(self.unit)
        cgs_unit = u_unit.cgs
        new_unit = pu_unit * cgs_unit / cgs_unit.scale
        return self._new_view(self.value * cgs_unit.scale, new_unit)

    @property
    def si(self):
        """
        Returns a copy of the current `PaicosQuantity` instance with SI units.
        The value of the resulting object will be scaled.
        """
        u_unit, pu_unit = separate_units(self.unit)
        si_unit = u_unit.si
        new_unit = pu_unit * si_unit / si_unit.scale
        return self._new_view(self.value * si_unit.scale, new_unit)

    def to(self, unit, equivalencies=[], copy=True):
        """
        Convert to different units. Similar functionality to the astropy
        Quantity.to() method.
        """
        if isinstance(unit, str):
            unit = u.Unit(unit)

        u_unit, pu_unit = separate_units(self.unit)

        u_unit_to, pu_unit_to = separate_units(unit)

        if pu_unit_to == u.Unit(''):
            return super().to(unit * pu_unit, equivalencies, copy)
        elif pu_unit == pu_unit_to:
            return super().to(unit, equivalencies, copy)
        else:
            err_msg = ('\n\nYou have requested conversion from\n {} '
                       + '\nto\n {} . \nThis is not possible as the a and h '
                       + 'factors differ. I.e. you cannot convert from\n '
                       + '{}\nto\n {}\nUse the .to_physical or .no_small_h '
                       + 'methods if you are trying to get rid of the a '
                       + 'and h factors.')

            raise RuntimeError(err_msg.format(self.unit, unit, pu_unit, pu_unit_to))

    @property
    def arepo(self):
        """
        Returns the quantity in Arepo code units.
        """
        arepo_bases = set([u.Unit('arepo_mass'),
                           u.Unit('arepo_length'),
                           u.Unit('arepo_time')])
        try:
            return self.decompose(bases=arepo_bases)
        except UnitConversionError as inst:
            err_msg = ('Conversion to arepo_units does not work well for '
                       + 'Temperature and magnetic field strength in Gauss. '
                       + 'Astropy throws the following error: ' + str(inst))

            raise UnitConversionError(err_msg) from inst

        return None

    @property
    def astro(self):
        """
        Returns the quantity in typical units used in cosmological simulations
        """
        return self.decompose(bases=[u.kpc, u.Msun, u.s, u.uG, u.keV, u.K])

    def decompose(self, bases=[]):
        """
        Decompose into a different set of units, e.g.

        A = B.decompose(bases=[u.kpc, u.Msun, u.s, u.uG, u.keV])

        small_a and small_h are automatically included in the bases.

        :meta private:
        """
        _, pu_unit = separate_units(self.unit)
        if len(bases) == 0 or pu_unit == u.Unit(''):
            return super().decompose(bases)

        if isinstance(bases, set):
            bases = list(bases)
        bases.append(small_a)
        bases.append(small_h)
        bases = set(bases)
        return super().decompose(bases)

    def label(self, variable=''):
        """
        Return a Latex string for use in plots. The optional
        input variable could be the Latex symbol for the physical variable,
        for instance \\rho or \\nabla\\times\\vec{v}.
        """

        a_sc, a_sc_str = self.__scaling_and_scaling_str(small_a)
        h_sc, h_sc_str = self.__scaling_and_scaling_str(small_h)

        co_label = a_sc_str + h_sc_str

        normal_unit = get_new_unit(self.unit, [small_h, small_a])

        _, dic = get_unit_dictionaries(self.unit)

        unit_label = ''
        for ii, key in enumerate(dic.keys()):
            _, unit_str = self.__scaling_and_scaling_str(key)
            unit_label += unit_str
            if ii < len(dic.keys()) - 1:
                unit_label += r'\;'

        label = co_label + r'\; \left[' + unit_label + r'\right]'

        # Get ckpc, cMpc, ckpc/h and Mkpc/h as used in literature
        if normal_unit in ('kpc', 'Mpc', 'Gpc'):
            if a_sc in (0, 1):
                if h_sc in (0, -1):
                    if a_sc == 1:
                        label = r'\mathrm{c}' + unit_label
                    elif a_sc == 0:
                        label = unit_label
                    if h_sc == -1:
                        label = label + r'\;h^{-1}'

            label = '[' + label + ']'

        if len(variable) > 0:
            label = variable + r'\;' + label
        label = '$' + label + '$'

        return label

    @property
    def to_physical(self):
        """
        Returns a copy of the current `PaicosQuantity` instance with the
        a and h factors removed, i.e. transform from comoving to physical.
        The value of the resulting object is scaled accordingly.
        """
        codic, _ = get_unit_dictionaries(self.unit)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = get_new_unit(self.unit, [small_a, small_h])
        return self._new_view(value * factor, new_unit)

    def to_comoving(self, unit):
        """
        Returns a copy of the current `PaicosQuantity` instance with the
        a and h factors given by the input unit.
        """

        if not self.comoving_sim:
            raise RuntimeError('Only implemented for comoving simulations')

        if isinstance(unit, str):
            unit = u.Unit(unit)

        u_unit_to, pu_unit_to = separate_units(unit)

        u_unit, pu_unit = separate_units(self.unit)

        if u_unit_to == u.Unit(''):
            u_unit_to = u_unit
        elif u_unit_to != u_unit:
            raise RuntimeError('This method can only change the a and h factors!')

        change_pu = pu_unit / pu_unit_to

        codic, _ = get_unit_dictionaries(change_pu)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = self.unit / change_pu

        return self._new_view(value * factor, new_unit)

    def to_any(self, new_unit):
        """
        Convert quantity to any (possible) unit.

        The .to method only allows converting standard
        physical units while the .to_comoving only allows
        modifying the a and h factors. This method allows
        changing the units in one go.
        """
        if isinstance(new_unit, str):
            new_unit = u.Unit(new_unit)

        u_unit, pu_unit = separate_units(new_unit)
        new_quant = self.to(u_unit)
        new_quant = new_quant.to_comoving(pu_unit)
        return new_quant

    def adjust_time(self, a=None, z=None):
        """
        Return a PaicosQuantity with its time (i.e. scale factor, a, and
        redshift, z, set to new values *without* changing the physical
        value of the quantity).

        Note the difference between this method and the "adjust_time" method!

        Use the "adjust_time" method if you want to keep the physical value fixed.
        Use the "set_time" method if you want to keep the comoving value fixed.

        Both are illustrated in the example below:

        For instance, we initialize a length of 100 comoving kpc
        (using a snap object with a = 1)

            In [2]:  L = 100 * snap.uq('kpc / small_a')
               ...:  L_fixed_physical = L.adjust_time(a=0.5)
               ...:  L_fixed_comoving = L.set_time(a=0.5)

            In [3]: L_fixed_physical
            Out[3]: <PaicosQuantity 50. kpc / small_a>

            In [4]: L_fixed_comoving
            Out[4]: <PaicosQuantity 100. kpc / small_a>

            In [5]: L.to_physical
            Out[5]: <PaicosQuantity 100. kpc>

            In [6]: L_fixed_physical.to_physical
            Out[6]: <PaicosQuantity 100. kpc>

            In [7]: L_fixed_comoving.to_physical
            Out[7]: <PaicosQuantity 200. kpc>

        """

        if not self.comoving_sim:
            raise RuntimeError('Only implemented for comoving simulations')

        codic, _ = get_unit_dictionaries(self.unit)
        # factor = self.h**codic[small_h]

        if a is not None:
            new_a = a
        elif z is not None:
            new_a = 1 / (z + 1)
        else:
            raise RuntimeError('You need to give an input to this method')

        factor = (self.a / new_a)**codic[small_a]
        data = self.value * factor
        return PaicosQuantity(data, self.unit, a=new_a, h=self.h,
                              comoving_sim=self.comoving_sim,
                              dtype=data.dtype, copy=True)

    def set_time(self, a=None, z=None):
        """
        Return a PaicosQuantity with its time (i.e. scale factor, a, and
        redshift, z, set to new values *while changing* the physical
        value of the quantity).

        Note the difference between this method and the "adjust_time" method!

        Use the "adjust_time" method if you want to keep the physical value fixed.
        Use the "set_time" method if you want to keep the comoving value fixed.

        Both are illustrated in the example below:

        For instance, we initialize a length of 100 comoving kpc
        (using a snap object with a = 1)

            In [2]:  L = 100 * snap.uq('kpc / small_a')
               ...:  L_fixed_physical = L.adjust_time(a=0.5)
               ...:  L_fixed_comoving = L.set_time(a=0.5)

            In [3]: L_fixed_physical
            Out[3]: <PaicosQuantity 50. kpc / small_a>

            In [4]: L_fixed_comoving
            Out[4]: <PaicosQuantity 100. kpc / small_a>

            In [5]: L.to_physical
            Out[5]: <PaicosQuantity 100. kpc>

            In [6]: L_fixed_physical.to_physical
            Out[6]: <PaicosQuantity 100. kpc>

            In [7]: L_fixed_comoving.to_physical
            Out[7]: <PaicosQuantity 200. kpc>

        """
        if not self.comoving_sim:
            raise RuntimeError('Only implemented for comoving simulations')

        if a is not None:
            new_a = a
        elif z is not None:
            new_a = 1 / (z + 1)
        else:
            raise RuntimeError('You need to give an input to this method')

        data = self.value
        return PaicosQuantity(data, self.unit, a=new_a, h=self.h,
                              comoving_sim=self.comoving_sim,
                              dtype=data.dtype, copy=True)

    @property
    def astropy_quantity(self):
        """
        Returns an astropy Quantity, i.e., the units are retained but
        we no longer have access to the information about the values of a and h.
        """
        return Quantity(self.value, self.unit, dtype=self.value.dtype, copy=True)

    def __scaling_and_scaling_str(self, unit):
        """
        Helper function to create labels

        :meta private:
        """
        codic, dic = get_unit_dictionaries(self.unit)
        # print(unit, dic, codic)
        if unit in codic:
            scaling = codic[unit]
        elif unit in dic:
            scaling = dic[unit]
        else:
            raise RuntimeError('should not happen')

        if unit in codic:
            base_string = unit.to_string(format='unicode')
        elif unit in dic:
            base_string = unit.to_string(format='latex')[1:-1]
        scaling_str = str(Fraction(scaling).limit_denominator(10000))
        if scaling_str == '0':
            scaling_str = ''
        elif scaling_str == '1':
            scaling_str = base_string
        else:
            scaling_str = base_string + '^{' + scaling_str + '}'
        return scaling, scaling_str

    def _sanity_check(self, value):
        """
        Function for sanity-checking addition, subtraction, multiplication
        and division of quantities. They should all have same a and h.
        """
        # pylint: disable=protected-access
        err_msg = "Operation requires objects to have same a and h value."
        if isinstance(value, PaicosQuantity):
            if value._a != self._a:
                info = f' Obj1._a={self._a}, Obj2._a={value._a}'
                raise RuntimeError(err_msg + info)
            if value.h != self.h:
                info = f' Obj1.h={self.h}, Obj2.h={value.h}'
                raise RuntimeError(err_msg + info)

    def _repr_latex_(self):
        number_part = super()._repr_latex_().split('\\;')[0]
        _, pu_units = separate_units(self.unit)
        u_latex = (self.unit / pu_units).to_string(format='latex')[1:-1]
        pu_latex = pu_units.to_string(format='latex')[1:-1]

        if pu_units != u.Unit(''):
            modified = number_part + '\\;' + u_latex + \
                '\\times' + pu_latex + '$'
        else:
            modified = number_part + '\\;' + u_latex + '$'
        return modified

    def __add__(self, value):
        self._sanity_check(value)
        return super().__add__(value)

    def __sub__(self, value):
        self._sanity_check(value)
        return super().__sub__(value)

    def __mul__(self, value):
        self._sanity_check(value)
        return super().__mul__(value)

    def __truediv__(self, value):
        self._sanity_check(value)
        return super().__truediv__(value)

    def dump(self):
        """
        This astropy Quantity method has not been implemented in the Paicos subclass.

        :meta private:
        """
        raise RuntimeError("not implemented")

    def dumps(self):
        """
        This astropy Quantity method has not been implemented in the Paicos subclass.

        :meta private:
        """
        raise RuntimeError("not implemented")

    def tobytes(self):
        """
        This astropy Quantity method has not been implemented in the Paicos subclass.

        :meta private:
        """
        raise RuntimeError("not implemented")

    def tofile(self):
        """
        This astropy Quantity method has not been implemented in the Paicos subclass.

        :meta private:
        """
        raise RuntimeError("not implemented")

    def tolist(self):
        """
        This astropy Quantity method has not been implemented in the Paicos subclass.

        :meta private:
        """
        raise RuntimeError("not implemented")

    def tostring(self):
        """
        This astropy Quantity method has not been implemented in the Paicos subclass.

        :meta private:
        """
        raise RuntimeError("not implemented")

    def choose(self):
        """
        This astropy Quantity method has not been implemented in the Paicos subclass.

        :meta private:
        """
        raise RuntimeError("not implemented")


class PaicosTimeSeries(PaicosQuantity):

    """
    PaicosTimeSeries is a subclass of the PaicosQuantity and and shares many
    of the same methods.

    The time series is very similar to the PaicosQuantity but
    the .a or .time properties now return arrays with same length
    as the object itself (.shape[0]).
    """

    def __new__(cls, value, unit=None, dtype=None, copy=True, order=None,
                subok=False, ndmin=0, h=None, a=None, comoving_sim=None):

        if isinstance(value, list):
            h = value[0].h  # Could check that they are all the same...
            comoving_sim = value[0].comoving_sim
            unit = value[0].unit
            dtype = value[0].dtype

            shapes = [value[i].value.shape for i in range(len(value))]
            err_msg = f'Not all shapes of the list are identical!, {shapes},\n\n'
            for shape in shapes:
                assert shape == shapes[0], err_msg + f'{shape}'

            a = np.array([value[i]._a for i in range(len(value))])
            value = np.array([value[i].value for i in range(len(value))])
        elif isinstance(value, np.ndarray):
            assert h is not None
            assert a is not None
            assert isinstance(a, np.ndarray)
            assert a.shape[0] == value.shape[0]
            if hasattr(value, 'unit'):
                if unit is not None:
                    raise RuntimeError(
                        'value has units but you are also passing unit={}:'.format(unit))
                unit = value.unit
                value = value.value
        else:
            raise RuntimeError('unexpected input for value:', value)

        msg = ('PaicosTimeSeries requires that the length of the first '
               + 'dimension is equal to the length of the time array')
        assert value.shape[0] == a.shape[0], msg

        assert len(value.shape) <= 3, 'Only 1D, 2D and 3D arrays are supported'

        obj = super().__new__(cls, value, unit=unit, dtype=dtype, copy=copy,
                              order=order, subok=subok, ndmin=ndmin, h=h, a=a,
                              comoving_sim=comoving_sim)

        obj._h = h
        obj._a = a
        obj._comoving_sim = comoving_sim

        return obj

    def __getitem__(self, key):
        """
        We overload this method in order to also slice the
        time array when a PaicosTimeSeries is sliced.
        """
        result = super().__getitem__(key)

        # Ensure key is always a tuple
        key_tuple = (key,) if not isinstance(key, tuple) else key

        # print(key_tuple)

        # Get index for axis 0
        key0 = key_tuple[0] if len(key_tuple) > 0 else slice(None)
        # print(key, key0)
        if not len(self._a.shape) == 0:
            result._a = self._a[key0]

        # Return PaicosQuantity if new view is just a single point in time
        if len(result._a.shape) == 0:
            return PaicosQuantity(result.value, result.unit, a=result.a,
                                  h=result.h, comoving_sim=result.comoving_sim,
                                  copy=True)

        return result

    def __array_function__(self, func, types, args, kwargs):
        """
        Ensure that the the information about a-values is correctly propagated.
        """

        result = super().__array_function__(func, types, args, kwargs)

        # Numpy functions that sometimes turn a time series into a single value
        if func in [np.mean, np.sum, np.std, np.var, np.min, np.max]:
            axis = kwargs.get("axis", None)
            if axis is None or (isinstance(axis, int) and axis == 0) or (isinstance(axis, tuple) and 0 in axis):
                # Reduction along dimension 0, so discard time information and return an astropy Quantity
                return result.astropy_quantity

        # Stacking functions
        if func in [np.vstack, np.hstack, np.dstack]:
            err_msg = ("This stacking function not implemented for PaicosTimeSeries. ",
                       "Please try instead if np.concatenate or np.stack works.")
            raise RuntimeError(err_msg)

        if func in [np.concatenate, np.stack]:

            arrays = args[0]
            if not all(isinstance(p, PaicosTimeSeries) for p in arrays):
                raise RuntimeError("concatenate requires same type for all arrays")

            axis = kwargs.get("axis", None)

            if axis == 0:
                if func == np.stack:
                    raise RuntimeError("np.stack only implemented with axis=1")
                # Concatenate the a-values if we are concatenating / stacking along axis 0
                new_a = np.concatenate([p.a for p in arrays])

                if len(new_a.shape) > 1:
                    print(new_a.shape)
                    err_msg = ("Supported operations for PaicosTimeSeries are np.concatenate",
                               " along axis=0 and/or stacking with np.stack along axis=1.")
                    raise RuntimeError(err_msg)
                return PaicosTimeSeries(result.value, result.unit, a=new_a,
                                        h=result.h, copy=True, comoving_sim=result.comoving_sim)
            elif axis == 1:
                a_first = arrays[0].a
                for p in arrays:
                    try:
                        np.testing.assert_array_equal(p.a, a_first)
                    except AssertionError:
                        err_msg = ("Stacking time series along this axis",
                                   " requires that all time series have identical",
                                   " time (i.e., .a values).")
                        raise RuntimeError(err_msg)
            else:
                raise RuntimeError(f"You have axis={axis}. Please specify axis to be either 0 or 1.")

        return result

    @property
    def to_physical(self):
        """
        Returns a copy of the current `PaicosTimeSeries` instance with the
        a and h factors removed, i.e. transform from comoving to physical.
        The value of the resulting object is scaled accordingly.
        """
        codic, _ = get_unit_dictionaries(self.unit)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = get_new_unit(self.unit, [small_a, small_h])

        if len(value.shape) == 1:
            new_value = value * factor
        elif len(value.shape) == 2:
            new_value = value * factor[:, None]
        elif len(value.shape) == 3:
            new_value = value * factor[:, None, None]

        return self._new_view(new_value, new_unit)

    def to_comoving(self, unit):
        """
        Returns a copy of the current `PaicosTimeSeries` with the
        a and h factors given by the input unit.
        """

        if not self.comoving_sim:
            raise RuntimeError('Only implemented for comoving simulations')

        if isinstance(unit, str):
            unit = u.Unit(unit)

        u_unit_to, pu_unit_to = separate_units(unit)
        u_unit, pu_unit = separate_units(self.unit)

        if u_unit_to == u.Unit(''):
            u_unit_to = u_unit
        elif u_unit_to != u_unit:
            raise RuntimeError('This method can only change the a and h factors!')

        change_pu = pu_unit / pu_unit_to

        codic, _ = get_unit_dictionaries(change_pu)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = self.unit / change_pu

        if len(value.shape) == 1:
            new_value = value * factor
        elif len(value.shape) == 2:
            new_value = value * factor[:, None]
        elif len(value.shape) == 3:
            new_value = value * factor[:, None, None]

        return self._new_view(new_value, new_unit)

    @property
    def hdf5_attrs(self):
        """
        Give the units as a dictionary for hdf5 data set attributes

        :meta private:
        """
        return {'unit': self.unit.to_string(), 'Paicos': 'PaicosTimeSeries'}

    @property
    def copy(self):
        """
        Returns a copy of the PaicosTimeSeries.
        """
        return PaicosTimeSeries(self.value, self.unit, a=self._a, h=self.h,
                                copy=True, comoving_sim=self.comoving_sim)

    def make_matrix(self, vec):
        """
        :meta private:
        """
        assert vec.shape[0] == self.shape[0]
        return np.vstack([vec for _ in range(self.shape[1])]).T

    # def concatenate_and_sort(self, paicos_time_series):
    #     """
    #     Concatenate two time series and sort according to their a values
    #     """
    #     assert self.unit == paicos_time_series.unit
    #     assert self.shape == paicos_time_series.shape

    #     # index = np.arange(self.a.shape[0] + paicos_time_series.a.shape[0])
    #     stacked_a = np.concatenate([self.a, paicos_time_series.a])
    #     stacked_values = np.concatenate([self.value, paicos_time_series.value])
    #     index = np.argsort(stacked_a)
    #     return PaicosTimeSeries(stacked_values[index], self.unit, a=stacked_a[index],
    #                             h=self.h, copy=True, comoving_sim=self.comoving_sim)

    def _sanity_check(self, value):
        """
        Function for sanity-checking addition, subtraction, multiplication
        and division of quantities. They should all have same a and h.
        """
        err_msg = "Operation requires objects to have same a and h value.\n"
        if isinstance(value, PaicosQuantity):
            if not isinstance(value, PaicosTimeSeries):
                msg = ('operations combining PaicosQuantity and '
                       + 'PaicosTimeSeries is not allowed.')
                raise RuntimeError(msg)

        if isinstance(value, PaicosTimeSeries):
            info = f'\nObj1.a={self._a}.\n\nObj2.a={value._a}'
            try:
                np.testing.assert_array_equal(value._a, self._a)
            except AssertionError as exc:
                raise RuntimeError(err_msg + info) from exc
            if value.h != self.h:
                info = f'\nObj1.h={self.h}.\n\nObj2.h={value.h}'
                raise RuntimeError(err_msg + info)


def paicos_quantity_list_to_array(list_to_convert):
    """
    Only works on 1D lists...

    :meta private:
    """
    e0 = list_to_convert[0]
    if hasattr(e0, 'unit'):
        for e in list_to_convert:
            assert e0.unit == e.unit
            assert e0.comoving_sim == e.comoving_sim
            assert e0._a == e._a
            assert e0._h == e._h
        return np.array([e.value for e in list_to_convert]) * e0.unit_quantity
    else:
        return np.array(list_to_convert)



import numpy as np
import numba

from .image_creator import ImageCreator
from .. import util
from .. import settings
from .. import units

from ..trees.bvh_cpu import BinaryTree
from ..trees.bvh_cpu import nearest_neighbor_cpu, nearest_neighbor_cpu_voronoi, nearest_neighbor_cpu_optimized

from ..trees.bvh_cpu import leading_zeros_cython_for_numba


@numba.njit(inline='always')
def part1by2_64(n):
    n &= 0x1fffff
    n = (n | (n << 32)) & 0x1f00000000ffff
    n = (n | (n << 16)) & 0x1f0000ff0000ff
    n = (n | (n << 8)) & 0x100f00f00f00f00f
    n = (n | (n << 4)) & 0x10c30c30c30c30c3
    n = (n | (n << 2)) & 0x1249249249249249
    return n


@numba.njit(inline='always')
def unpart1by2_64(n):
    n = numba.uint64(n)
    n &= 0x1249249249249249
    n = (n ^ (n >> 2)) & 0x10c30c30c30c30c3
    n = (n ^ (n >> 4)) & 0x100f00f00f00f00f
    n = (n ^ (n >> 8)) & 0x1f0000ff0000ff
    n = (n ^ (n >> 16)) & 0x1f00000000ffff
    n = (n ^ (n >> 32)) & 0x1fffff
    return n


@numba.njit(inline='always')
def encode64(x, y, z):
    xx = part1by2_64(x)
    yy = part1by2_64(y)
    zz = part1by2_64(z)
    return xx * 4 + yy * 2 + zz


@numba.njit(parallel=True)
def trace_rays_cpu(points, tree_parents, tree_children, tree_bounds, variable, hsml,
                   widths, center, tree_scale_factor, tree_offsets, image, rotation_matrix, tol):

    nx = image.shape[0]
    ny = image.shape[1]
    dx = widths[0] / nx
    dy = widths[1] / ny
    num_internal_nodes = tree_children.shape[0]

    for ix in numba.prange(nx):  # pylint: disable=not-an-iterable
        for iy in range(ny):
            result = 0.0
            z = 0.0
            query_point = np.empty(3, dtype=np.float64)

            while z < widths[2]:
                # Compute query point in aligned coords
                query_point[0] = (center[0] - widths[0]
                                  / 2.0) + (ix + 0.5) * dx
                query_point[1] = (center[1] - widths[1]
                                  / 2.0) + (iy + 0.5) * dy
                query_point[2] = (center[2] - widths[2] / 2.0) + z

                # Rotate point around center
                query_point = np.dot(
                    rotation_matrix, query_point - center) + center

                # Convert to tree coordinates
                query_point[0] = (
                    query_point[0] - tree_offsets[0]) * tree_scale_factor
                query_point[1] = (
                    query_point[1] - tree_offsets[1]) * tree_scale_factor
                query_point[2] = (
                    query_point[2] - tree_offsets[2]) * tree_scale_factor

                # Nearest neighbor search in tree coordinates
                min_dist, min_index = nearest_neighbor_cpu(points, tree_parents, tree_children,
                                                           tree_bounds, query_point,
                                                           num_internal_nodes)

                # Adaptive step in Arepo units
                # dz = tol * hsml[min_index]
                dz = tol * min_dist / tree_scale_factor
                # if hsml[min_index] < min_dist:
                # print(hsml[min_index], min_dist, min_dist/tree_scale_factor)
                result += dz * variable[min_index]
                z += dz

            # Correct overshoot
            result -= (z - widths[2]) * variable[min_index]
            image[ix, iy] = result


@numba.njit(parallel=True)
def trace_rays_cpu_voronoi(points, tree_parents, tree_children, tree_bounds, variable, hsml,
                           widths, center, tree_scale_factor, tree_offsets, image, rotation_matrix, tol):

    nx = image.shape[0]
    ny = image.shape[1]
    dx = widths[0] / nx
    dy = widths[1] / ny
    num_internal_nodes = tree_children.shape[0]
    L = 21

    for ix in numba.prange(nx):  # pylint: disable=not-an-iterable
        for iy in range(ny):
            result = 0.0
            z = 0.0
            query_point = np.empty(3, dtype=np.float64)

            while z < widths[2]:
                # Compute query point in aligned coords
                query_point[0] = (center[0] - widths[0]
                                  / 2.0) + (ix + 0.5) * dx
                query_point[1] = (center[1] - widths[1]
                                  / 2.0) + (iy + 0.5) * dy
                query_point[2] = (center[2] - widths[2] / 2.0) + z

                # Rotate point around center
                query_point = np.dot(
                    rotation_matrix, query_point - center) + center

                # Convert to tree coordinates
                query_point[0] = (
                    query_point[0] - tree_offsets[0]) * tree_scale_factor
                query_point[1] = (
                    query_point[1] - tree_offsets[1]) * tree_scale_factor
                query_point[2] = (
                    query_point[2] - tree_offsets[2]) * tree_scale_factor

                # Nearest neighbor search in tree coordinates
                min_dist, min_index = nearest_neighbor_cpu_voronoi(points, tree_parents, tree_children,
                                                                   tree_bounds, query_point,
                                                                   num_internal_nodes)
                if min_index == -1:
                    min_dist, min_index = nearest_neighbor_cpu_optimized(points, tree_parents, tree_children,
                                                                         tree_bounds, query_point,
                                                                         num_internal_nodes, (2**L - 1.0))
                # Adaptive step in Arepo units
                dz = tol * hsml[min_index]
                result += dz * variable[min_index]
                z += dz

            # Correct overshoot
            result -= (z - widths[2]) * variable[min_index]
            image[ix, iy] = result


@numba.njit(parallel=True)
def trace_rays_several_variables_cpu_voronoi(points, tree_parents, tree_children, tree_bounds,
                                             variables, hsml, widths, center,
                                             tree_scale_factor, tree_offsets, images,
                                             rotation_matrix, tol):

    nx = images.shape[0]
    ny = images.shape[1]
    nvars = images.shape[2]
    dx = widths[0] / nx
    dy = widths[1] / ny
    num_internal_nodes = tree_children.shape[0]
    L = 21

    for ix in numba.prange(nx):  # pylint: disable=not-an-iterable
        for iy in range(ny):
            result = np.zeros(nvars, dtype=variables.dtype)
            z = 0.0
            query_point = np.empty(3, dtype=np.float64)

            while z < widths[2]:
                # Compute query point in aligned coords
                query_point[0] = (center[0] - widths[0]
                                  / 2.0) + (ix + 0.5) * dx
                query_point[1] = (center[1] - widths[1]
                                  / 2.0) + (iy + 0.5) * dy
                query_point[2] = (center[2] - widths[2] / 2.0) + z

                # Rotate point around center
                query_point = np.dot(
                    rotation_matrix, query_point - center) + center

                # Convert to tree coordinates
                query_point[0] = (
                    query_point[0] - tree_offsets[0]) * tree_scale_factor
                query_point[1] = (
                    query_point[1] - tree_offsets[1]) * tree_scale_factor
                query_point[2] = (
                    query_point[2] - tree_offsets[2]) * tree_scale_factor

                # Nearest neighbor search in tree coordinates
                min_dist, min_index = nearest_neighbor_cpu_voronoi(points, tree_parents, tree_children,
                                                                   tree_bounds, query_point,
                                                                   num_internal_nodes)
                if min_index == -1:
                    min_dist, min_index = nearest_neighbor_cpu_optimized(points, tree_parents, tree_children,
                                                                         tree_bounds, query_point,
                                                                         num_internal_nodes, (2**L - 1.0))

                # Adaptive step in Arepo units
                dz = tol * hsml[min_index]

                for k in range(nvars):
                    result[k] += dz * variables[min_index, k]
                z += dz

            # Correct overshoot
            for k in range(nvars):
                result[k] -= (z - widths[2]) * variables[min_index, k]
                images[ix, iy, k] = result[k]


@numba.njit(parallel=True)
def trace_rays_cpu_optimized(points, tree_parents, tree_children, tree_bounds, variable, hsml,
                             widths, center, tree_scale_factor, tree_offsets, image, rotation_matrix, tol):

    nx = image.shape[0]
    ny = image.shape[1]
    dx = widths[0] / nx
    dy = widths[1] / ny
    num_internal_nodes = tree_children.shape[0]

    L = 21

    for ix in numba.prange(nx):  # pylint: disable=not-an-iterable
        for iy in range(ny):
            result = 0.0
            z = 0.0
            query_point = np.empty(3, dtype=np.float64)

            max_search_dist = (2**L - 1.0)

            while z < widths[2]:
                # Compute query point in aligned coords
                query_point[0] = (center[0] - widths[0]
                                  / 2.0) + (ix + 0.5) * dx
                query_point[1] = (center[1] - widths[1]
                                  / 2.0) + (iy + 0.5) * dy
                query_point[2] = (center[2] - widths[2] / 2.0) + z

                # Rotate point around center
                query_point = np.dot(
                    rotation_matrix, query_point - center) + center

                # Convert to tree coordinates
                query_point[0] = (
                    query_point[0] - tree_offsets[0]) * tree_scale_factor
                query_point[1] = (
                    query_point[1] - tree_offsets[1]) * tree_scale_factor
                query_point[2] = (
                    query_point[2] - tree_offsets[2]) * tree_scale_factor

                # Nearest neighbor search in tree coordinates
                min_dist, min_index = nearest_neighbor_cpu_optimized(points, tree_parents, tree_children,
                                                                     tree_bounds, query_point,
                                                                     num_internal_nodes, max_search_dist)
                # Adaptive step in Arepo units
                dz = 2.0 * tol * min_dist / tree_scale_factor
                max_search_dist = 1.1 * (min_dist + dz * tree_scale_factor)
                result += dz * variable[min_index]
                z += dz

            # Correct overshoot
            result -= (z - widths[2]) * variable[min_index]
            image[ix, iy] = result


@numba.njit(parallel=True)
def trace_rays_several_variables_cpu_optimized(points, tree_parents, tree_children, tree_bounds,
                                               variables, hsml, widths, center,
                                               tree_scale_factor, tree_offsets, images,
                                               rotation_matrix, tol):

    nx = images.shape[0]
    ny = images.shape[1]
    nvars = images.shape[2]
    dx = widths[0] / nx
    dy = widths[1] / ny
    num_internal_nodes = tree_children.shape[0]

    L = 21

    for ix in numba.prange(nx):  # pylint: disable=not-an-iterable
        for iy in range(ny):
            result = np.zeros(nvars, dtype=variables.dtype)
            z = 0.0
            query_point = np.empty(3, dtype=np.float64)

            max_search_dist = (2**L - 1.0)

            while z < widths[2]:
                # Compute query point in aligned coords
                query_point[0] = (center[0] - widths[0]
                                  / 2.0) + (ix + 0.5) * dx
                query_point[1] = (center[1] - widths[1]
                                  / 2.0) + (iy + 0.5) * dy
                query_point[2] = (center[2] - widths[2] / 2.0) + z

                # Rotate point around center
                query_point = np.dot(
                    rotation_matrix, query_point - center) + center

                # Convert to tree coordinates
                query_point[0] = (
                    query_point[0] - tree_offsets[0]) * tree_scale_factor
                query_point[1] = (
                    query_point[1] - tree_offsets[1]) * tree_scale_factor
                query_point[2] = (
                    query_point[2] - tree_offsets[2]) * tree_scale_factor

                # Nearest neighbor search in tree coordinates
                min_dist, min_index = nearest_neighbor_cpu_optimized(points, tree_parents, tree_children,
                                                                     tree_bounds, query_point,
                                                                     num_internal_nodes, max_search_dist)
                # Adaptive step in Arepo units
                dz = 2.0 * tol * min_dist / tree_scale_factor
                max_search_dist = 1.1 * (min_dist + dz * tree_scale_factor)

                for k in range(nvars):
                    result[k] += dz * variables[min_index, k]
                z += dz

            # Correct overshoot
            for k in range(nvars):
                result[k] -= (z - widths[2]) * variables[min_index, k]
                images[ix, iy, k] = result[k]


class RayProjector(ImageCreator):
    """
    A class that allows creating an image of a given variable by projecting
    it onto a 2D plane. This class works by raytracing the variable
    (i.e. by calculating a line integral along the line-of-sight).

    This is crudely made CPU-version which was made by adapting
    the corresponding GPU code back to CPU.
    """

    @util.conditional_timer
    def __init__(self, snap, center, widths, direction,
                 npix=512, parttype=0, tol=0.25, do_pre_selection=False,
                 use_numba=True, verbose=False, timing=False):
        """
        Initialize the Projector class.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center : numpy array
            Center of the region on which projection is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths : numpy array
            Widths of the region on which projection is to be done,
            e.g.m widths=[width_x, width_y, width_z].

        direction : str
            Direction of the projection, e.g. 'x', 'y' or 'z',
            or a Paicos Orientation class instance.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        parttype : int, optional
            Number of the particle type to project, by default gas (PartType 0).

        """

        # call the superclass constructor to initialize the ImageCreator class
        super().__init__(snap, center, widths, direction, npix=npix, parttype=parttype)

        parttype = self.parttype

        self.do_pre_selection = do_pre_selection

        self.tol = tol

        self.use_numba = use_numba

        self.verbose = verbose

        # Calculate the smoothing length
        avail_list = (list(snap.keys()) + snap._auto_list)
        if f'{parttype}_Volume' in avail_list:
            self.hsml = 1.2 * 2.0 * np.cbrt((self.snap[f"{parttype}_Volume"])
                                            / (4.0 * np.pi / 3.0))
        elif f'{parttype}_SubfindHsml' in avail_list:
            self.hsml = self.snap[f'{parttype}_SubfindHsml']
        else:
            if self.parttype != 0:
                self.hsml = np.zeros_like(self.snap[f'{parttype}_Coordinates'])
            else:
                raise RuntimeError(
                    'There is no smoothing length or volume for the projector')

        self.pos = self.snap[f'{self.parttype}_Coordinates']

        if settings.use_units:
            self.hsml = self.hsml.to(self.pos.unit)

        # Call selection
        self.has_do_region_selection_been_called = False
        self._do_region_selection()
        self.has_do_region_selection_been_called = True

    def _do_region_selection(self):

        self.do_unit_consistency_check()

        center = self.center
        widths = self.widths
        snap = self.snap
        parttype = self.parttype

        if self.has_do_region_selection_been_called:
            if self.do_pre_selection:
                import warnings
                err_msg = ("It looks like you are changing projector "
                           + " properties after the fact, i.e. changing widths "
                           + "center, orientation, resolution etc. This might be "
                           + "slow with with the option do_pre_selection, which "
                           + "you have turned on. If you have enough memory "
                           + "then it is probably better to set do_pre_selection "
                           + "to False.")
                warnings.warn(err_msg)

        # Send subset of snapshot to GPU
        if self.do_pre_selection:
            # get the index of the region of projection
            if self.direction != 'orientation':
                get_index = util.get_index_of_cubic_region_plus_thin_layer
                self.index = get_index(self.snap[f"{parttype}_Coordinates"],
                                       center, widths, self.hsml, snap.box)
            else:
                get_index = util.get_index_of_rotated_cubic_region_plus_thin_layer
                self.index = get_index(snap[f"{parttype}_Coordinates"],
                                       center, widths, self.hsml, snap.box,
                                       self.orientation)

            self.hsml = self.hsml[self.index]
            self.pos = self.pos[self.index]

            self._send_data_to_tree()

            # We need to reconstruct the tree!
            self.tree = BinaryTree(self.tree_variables['pos'],
                                   self.tree_variables['hsml'], self.verbose)
            del self.tree_variables['pos']
            self.tree_variables['hsml'] = self.tree_variables['hsml'][self.tree.sort_index]
        # Send entirety of snapshot to GPU (if we have not already
        # done so). Always send small data with change in resolution etc
        else:
            if not self.has_do_region_selection_been_called:
                self._send_data_to_tree()

                # Construct tree
                self.tree = BinaryTree(self.tree_variables['pos'],
                                       self.tree_variables['hsml'], self.verbose)

                del self.tree_variables['pos']
                self.tree_variables['hsml'] = self.tree_variables['hsml'][
                    self.tree.sort_index]

        # Always send small data
        self._send_small_data_to_tree()

    def _send_data_to_tree(self):
        self.tree_variables = {}
        if settings.use_units:
            self.tree_variables['pos'] = np.array(self.pos.value)
            self.tree_variables['hsml'] = np.array(self.hsml.value)
        else:
            self.tree_variables['pos'] = np.array(self.pos)
            self.tree_variables['hsml'] = np.array(self.hsml)

        self._send_small_data_to_tree()

    def _send_small_data_to_tree(self):

        self.tree_variables['rotation_matrix'] = np.array(
            self.orientation.rotation_matrix)

        if settings.use_units:
            self.tree_variables['widths'] = np.array(self.widths.value)
            self.tree_variables['center'] = np.array(self.center.value)
        else:
            self.tree_variables['widths'] = np.array(self.widths)
            self.tree_variables['center'] = np.array(self.center)

    def _tree_project(self, variable_str, additive):
        """
        Private method for projecting using cuda code
        """
        tree_vars = self.tree_variables
        rotation_matrix = tree_vars['rotation_matrix']
        widths = tree_vars['widths']
        center = tree_vars['center']
        hsml = tree_vars['hsml']
        nx = self.npix_width
        ny = self.npix_height

        if additive:
            variable = tree_vars[variable_str] / tree_vars[f'{self.parttype}_Volume']
        else:
            variable = tree_vars[variable_str]

        tree_scale_factor = self.tree.conversion_factor
        tree_offsets = self.tree.off_sets

        image = np.zeros((nx, ny))

        if self.use_numba:

            if self.parttype == 0:
                trace_rays = trace_rays_cpu_voronoi
            else:
                trace_rays = trace_rays_cpu_optimized

            trace_rays(self.tree._pos, self.tree.parents, self.tree.children,
                       self.tree.bounds, variable, hsml, widths, center,
                       tree_scale_factor, tree_offsets, image,
                       rotation_matrix, self.tol)
        else:
            import warnings
            warnings.warn(
                "Cython ray tracer slower and less well tested than Numba!")
            if self.parttype == 0:
                from ..cython.ray_tracer import trace_rays_cpu_voronoi as trace_rays_cpu_cython
            else:
                from ..cython.ray_tracer import trace_rays_cpu as trace_rays_cpu_cython

            trace_rays_cpu_cython(self.tree._pos, self.tree.parents, self.tree.children,
                                  self.tree.bounds, variable, hsml, widths, center,
                                  tree_scale_factor, tree_offsets, image,
                                  rotation_matrix, self.tol, 1)

        return image

    def _tree_project_several_variables(self, variable_strs, additives):
        """
        Private method for projecting several variables in one tree traversal.

        Parameters
        ----------
        variable_strs : list of str
            Keys in `self.tree_variables` for each variable to project.
        additives : list of bool
            Whether to use additive projection for each variable.

        Returns
        -------
        numpy.ndarray
            Array of shape (nx, ny, n_variables) with the projected results.
        """
        tree_vars = self.tree_variables
        rotation_matrix = tree_vars['rotation_matrix']
        widths = tree_vars['widths']
        center = tree_vars['center']
        hsml = tree_vars['hsml']
        nx = self.npix_width
        ny = self.npix_height

        # Build (N, n_variables) array for particle variables
        var_list = []
        for var_str, additive in zip(variable_strs, additives):
            if additive:
                var_arr = tree_vars[var_str] / tree_vars[f'{self.parttype}_Volume']
            else:
                var_arr = tree_vars[var_str]
            var_list.append(var_arr)

        variables = np.column_stack(var_list)  # shape: (N, n_variables)

        tree_scale_factor = self.tree.conversion_factor
        tree_offsets = self.tree.off_sets

        # Allocate output cube (nx, ny, n_variables)
        images = np.zeros((nx, ny, len(variable_strs)), dtype=variables.dtype)

        if self.use_numba:
            if self.parttype == 0:
                trace_rays = trace_rays_several_variables_cpu_voronoi
            else:
                trace_rays = trace_rays_several_variables_cpu_optimized

            trace_rays(self.tree._pos, self.tree.parents, self.tree.children,
                       self.tree.bounds, variables, hsml, widths, center,
                       tree_scale_factor, tree_offsets, images,
                       rotation_matrix, self.tol)
        else:
            raise RuntimeError("only implemented for numba")

        return images

    def _send_variable_to_tree(self, variable, tree_key='projection_variable'):
        if isinstance(variable, str):
            variable_str = str(variable)
            err_msg = 'projector uses a different parttype'
            assert int(variable[0]) == self.parttype, err_msg
            variable = self.snap[variable]
        else:
            variable_str = tree_key
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        assert len(variable.shape) == 1, 'only scalars can be projected'

        # Select same part of array that the projector has selected
        if self.do_pre_selection:
            # TODO: Check that this is not applied more than once...
            variable = variable[self.index]

        if variable_str in self.tree_variables and variable_str != tree_key:
            pass
        else:
            # Send variable to gpu
            if settings.use_units:
                self.tree_variables[variable_str] = np.array(variable.value)
            else:
                self.tree_variables[variable_str] = np.array(variable)

            # Sort the variable according to Morton code sorting
            self.tree_variables[variable_str] = self.tree_variables[variable_str][
                self.tree.sort_index]

        if isinstance(variable, units.PaicosQuantity):
            unit_quantity = variable.unit_quantity
        else:
            unit_quantity = None

        return variable_str, unit_quantity

    @util.conditional_timer
    def project_variable(self, variable, additive=False, timing=False):
        """
        projects a given variable onto a 2D plane.

        Parameters
        ----------
        variable : str, function, numpy array
            variable, it can be passed as string or an array

        Returns
        -------
        numpy array
            The image of the projected variable
        """

        # This calls _do_region_selection if resolution, Orientation,
        # widths or center changed
        self._check_if_properties_changed()

        variable_str, unit_quantity = self._send_variable_to_tree(variable)
        if additive:
            _, vol_unit_quantity = self._send_variable_to_tree(f'{self.parttype}_Volume')

        # Do the projection
        projection = self._tree_project(variable_str, additive)

        # Transpose
        projection = projection.T

        assert projection.shape[0] == self.npix_height
        assert projection.shape[1] == self.npix_width

        if unit_quantity is not None:
            unit_length = self.snap['0_Coordinates'].uq
            projection = projection * unit_quantity * unit_length
            if additive:
                projection = projection / vol_unit_quantity

        if additive:
            return projection
        else:
            return projection / self.depth

    @util.conditional_timer
    def project_variables(self, variables, additives, timing=False):
        """
        Projects multiple variables onto a 2D plane in a single tree traversal.

        This probably offers a speed benefit compared to calling
        project_variable several times.

        Parameters
        ----------
        variables : list of str, function, or numpy array
            Variables to project.
        additives : list of bool
            Whether to use additive projection for each variable.

        Returns
        -------
        list
            List of 2D arrays, each being the projection of the corresponding variable.
        """
        assert len(variables) == len(additives), \
            "Length of `variables` and `additives` must match."

        self._check_if_properties_changed()

        tree_keys = []
        unit_quantities = []
        vol_unit_quantities = []
        is_additive_flags = []

        for i, (variable, additive) in enumerate(zip(variables, additives)):
            # Assign unique tree key for this projection variable
            # (not used for string variables!)
            tree_key = f'projection_variable{i}'

            variable_str, unit_quantity = self._send_variable_to_tree(
                variable, tree_key=tree_key
            )
            tree_keys.append(variable_str)
            unit_quantities.append(unit_quantity)
            is_additive_flags.append(additive)

            if additive:
                _, vol_unit_quantity = self._send_variable_to_tree(
                    f'{self.parttype}_Volume'
                )
            else:
                vol_unit_quantity = None
            vol_unit_quantities.append(vol_unit_quantity)

        # Perform projection in one tree traversal
        projection_cube = self._tree_project_several_variables(
            tree_keys, is_additive_flags)
        # projection_cube.shape = (nx, ny, n_variables)

        results = []
        for i in range(len(variables)):
            # transpose for output convention
            proj_2d = projection_cube[:, :, i].T

            assert proj_2d.shape[0] == self.npix_height
            assert proj_2d.shape[1] == self.npix_width

            unit_quantity = unit_quantities[i]
            vol_unit_quantity = vol_unit_quantities[i]
            additive = is_additive_flags[i]

            if unit_quantity is not None:
                unit_length = self.snap['0_Coordinates'].uq
                proj_2d = proj_2d * unit_quantity * unit_length
                if additive:
                    proj_2d = proj_2d / vol_unit_quantity

            if not additive:
                proj_2d = proj_2d / self.depth

            results.append(proj_2d)

        return results

    def __del__(self):
        """
        Clean up like this? Not sure it is needed...
        """
        self.release_tree_memory()

    def release_tree_memory(self):
        if hasattr(self, 'tree_variables'):
            for key in list(self.tree_variables):
                del self.tree_variables[key]
            del self.tree_variables
        if hasattr(self, 'tree'):
            # self.tree.release_tree_memory()
            del self.tree



"""This defines a reader for Arepo Friends-of-Friends (FoF) and subhalo catalog files"""
import numpy as np
import h5py
from .paicos_readers import PaicosReader
from ..writers.paicos_writer import PaicosWriter
from .. import settings
import numbers
import warnings


class PaicosDict(dict):
    def __init__(self, cat, subfind_catalog=True):

        self.cat = cat
        group_keys, sub_keys = self.cat.info(verbose=False)
        if subfind_catalog:
            self.loader = self.cat.load_sub_data
            self._auto_list = sub_keys
        else:
            self.loader = self.cat.load_group_data
            self._auto_list = group_keys

    def __getitem__(self, ikey):
        self.loader(ikey)
        return super().__getitem__(ikey)

    def _ipython_key_completions_(self):
        """
        Auto-completion of dictionary.
        """

        return self._auto_list


class Catalog(PaicosReader):
    """
    This is a Python class for reading Arepo group and subhalo catalogs.
    The class is based on a script originally written by Ewald Puchwein,
    which has since then been modified and included in Paicos.

    The class takes in the path of the directory containing the catalog, the
    catalog number, and an optional basename parameter, and uses this
    information to locate and open the catalog files. The class also loads
    the catalog's header, parameters, and configuration.
    The class also includes methods to extract the redshift, scale
    factor, and other properties of the catalog.

    Important methods and attributes:
    ---------------------------------

        cat = Catalog()

        cat.Group : dict
            Contains a dictionary of the FoF-catalog.

        cat.Sub : dict
            Contains a dictionary of the Subfind-catalog.

        cat.Parameters : dict
            Contains information from the parameter file used in the simulation (e.g. param.txt).

        cat.Config : dict
            Contains information from the Config file used in the simulation (e.g. Config.txt).

        cat.Header : dict
            Contains information about this particular catalog such as its time (e.g scale factor).

        cat.z : float
            The redshift.

        cat.h : float
            Reduced Hubble param (e.g. 0.67).

        cat.age : float
            The age of the Universe (only for cosmological runs).

        cat.lookback_time : float
            The age of the Universe (only for cosmological runs).

    """
    def __init__(self, basedir='.', snapnum=None, load_all=False,
                 to_physical=False, subfind_catalog=True, readonly_first_file=False,
                 verbose=False):
        """
        Initializes the Catalog class.

        Parameters
        ----------

            basedir : str
                The path of the directory containing the catalogs
                (e.g. the 'output' folder).

            snapnum : int
                The snapshot number.

            load_all : bool
                Whether to immediately load all fields or not.

            to_physical : bool
                whether to convert to physical unit upon loading the data.
                Default is False.

            subfind_catalog : bool
                whether the simulation has subfind catalogs,
                when False the code will look for FoF-catalogs only.

            readonly_first_file : bool
                when set to True the code will only read one catalog file (the first).
                This can significantly speed things up when one is only interested in the
                properties of the most massive groups in a simulation.

            verbose : bool
                whether to print information, default is False.

        """

        if subfind_catalog:
            basename = 'fof_subhalo_tab'
        else:
            basename = 'fof_tab'

        super().__init__(basedir=basedir, snapnum=snapnum, basename=basename,
                         load_all=False, to_physical=to_physical,
                         basesubdir='groups', verbose=verbose)

        # give names to specific header fields
        self.nfiles = self.Header["NumFiles"]
        self.ngroups = self.Header["Ngroups_Total"]

        if "Nsubgroups_Total" in self.Header.keys():
            self.nsubs = self.Header["Nsubgroups_Total"]
        else:
            self.nsubs = self.Header["Nsubhalos_Total"]

        # Initialize dictionaries
        self.Group = {}
        self.Sub = {}

        self.Group = PaicosDict(self, subfind_catalog=False)
        self.Sub = PaicosDict(self, subfind_catalog=True)

        self.to_physical = to_physical

        self.readonly_first_file = readonly_first_file

        if to_physical and not settings.use_units:
            err_msg = "to_physical=True requires that units are enabled"
            raise RuntimeError(err_msg)

        # Load all data
        if load_all:
            self.load_all_data()

    def info(self, verbose=True):
        """
        Return the available keys for the FoF catalog and/or
        the subhalo catalo.

        :meta private:
        """
        ifile = 0
        if self.multi_file is False:
            cur_filename = self.filename
        else:
            if self.no_subdir:
                cur_filename = self.multi_wo_dir.format(ifile)
            else:
                cur_filename = self.multi_filename.format(ifile)

        if self.verbose:
            print("reading file", cur_filename)

        file = h5py.File(cur_filename, "r")

        if "Group" in file:
            group_keys = list(file["Group"].keys())
        else:
            group_keys = []

        if "Subhalo" in file:
            sub_keys = list(file["Subhalo"].keys())
        else:
            sub_keys = []

        file.close()

        if not settings.use_units:
            if verbose:
                if len(group_keys) > 0:
                    print('Available group_keys are:', group_keys)

                if len(sub_keys) > 0:
                    print('\nAvailable sub_keys are:', sub_keys)
            return group_keys, sub_keys
        else:
            from .. import unit_specifications

        not_implemented_group_keys = []
        implemented_group_keys = []
        for key in list(group_keys):
            if key in unit_specifications.unit_dict['groups']:
                implemented_group_keys.append(key)
            else:
                not_implemented_group_keys.append(key)

        not_implemented_sub_keys = []
        implemented_sub_keys = []
        for key in list(sub_keys):
            if key in unit_specifications.unit_dict['subhalos']:
                implemented_sub_keys.append(key)
            else:
                not_implemented_sub_keys.append(key)

        if verbose:
            if len(group_keys) > 0:
                print('Available group_keys are:', implemented_group_keys)
                if len(not_implemented_group_keys) > 0:
                    print('\nNot implemented group_keys are:', not_implemented_group_keys)

            if len(sub_keys) > 0:
                print('\nAvailable sub_keys are:', implemented_sub_keys)
                if len(not_implemented_sub_keys) > 0:
                    print('\nNot implemented sub_keys are:', not_implemented_sub_keys)

        return implemented_group_keys, implemented_sub_keys

    def load_data(self):
        """
        Overwrite of the base class method.

        :meta private:
        """
        raise RuntimeError('should not be called')

    def load_group_data(self, ikey):
        """
        Load subhalo dato. See self.info for valid ikeys

        :meta private:
        """

        if ikey in self.Group:
            return

        data = None  # to get rid of linting error

        if not self.readonly_first_file:
            skip_gr = 0

            for ifile in range(self.nfiles):
                if self.multi_file is False:
                    cur_filename = self.filename
                else:
                    if self.no_subdir:
                        cur_filename = self.multi_wo_dir.format(ifile)
                    else:
                        cur_filename = self.multi_filename.format(ifile)

                if self.verbose:
                    print("reading file", cur_filename)

                file = h5py.File(cur_filename, "r")

                ng = int(file["Header"].attrs["Ngroups_ThisFile"])

                # initialize arrays
                if ifile == 0:
                    if "Group" in file:
                        if len(file["Group/" + ikey].shape) == 1:
                            data = np.empty(
                                self.ngroups, dtype=file["Group/" + ikey].dtype)
                        elif len(file["Group/" + ikey].shape) == 2:
                            data = np.empty(
                                (self.ngroups, file["Group/" + ikey].shape[1]),
                                dtype=file["Group/" + ikey].dtype)
                        else:
                            assert False

                # read group data
                if ng > 0:
                    data[skip_gr:skip_gr + ng] = file["Group/" + ikey]

                skip_gr += ng

                file.close()
        else:
            cur_filename = self.filename
            file = h5py.File(cur_filename, "r")
            ng = int(file["Header"].attrs["Ngroups_ThisFile"])
            if "Group" in file:
                if len(file["Group/" + ikey].shape) == 1:
                    data = np.empty(
                        ng, dtype=file["Group/" + ikey].dtype)
                elif len(file["Group/" + ikey].shape) == 2:
                    data = np.empty(
                        (ng, file["Group/" + ikey].shape[1]),
                        dtype=file["Group/" + ikey].dtype)
                else:
                    assert False

            # read group data
            if ng > 0:
                data[:ng] = file["Group/" + ikey][...]

            file.close()

        # Load all variables with double precision
        if settings.double_precision:
            if not issubclass(data.dtype.type, numbers.Integral):
                data = data.astype(np.float64)

        else:
            warnings.warn('\n\nThe cython routines expect double precision '
                          + 'and will fail unless settings.double_precision '
                          + 'is True.\n\n')

        if settings.use_units:
            self.Group[ikey] = self.get_paicos_quantity(
                data, ikey,
                field='groups')
            if not hasattr(self.Group[ikey], 'unit'):
                del self.Group[ikey]
                raise RuntimeError(f"{ikey} does not have units implemented!")
        else:
            self.Group[ikey] = data

    def load_sub_data(self, ikey):
        """
        Load subhalo dato. See self.info for valid ikeys

        TODO: Merge with load_group_data, lot's of duplicate code here...

        :meta private:
        """

        if ikey in self.Sub:
            return

        data = None  # to get rid of linting error

        skip_sub = 0
        for ifile in range(self.nfiles):
            if self.multi_file is False:
                cur_filename = self.filename
            else:
                if self.no_subdir:
                    cur_filename = self.multi_wo_dir.format(ifile)
                else:
                    cur_filename = self.multi_filename.format(ifile)

            if self.verbose:
                print("reading file", cur_filename)

            file = h5py.File(cur_filename, "r")

            if "Nsubgroups_ThisFile" in file["Header"].attrs.keys():
                ns = int(file["Header"].attrs["Nsubgroups_ThisFile"])
            else:
                ns = int(file["Header"].attrs["Nsubhalos_ThisFile"])

            # initialize arrays
            if ifile == 0:
                if "Subhalo" in file:
                    if len(file["Subhalo/" + ikey].shape) == 1:
                        data = np.empty(
                            self.nsubs, dtype=file["Subhalo/" + ikey].dtype)
                    elif len(file["Subhalo/" + ikey].shape) == 2:
                        data = np.empty(
                            (self.nsubs, file["Subhalo/" + ikey].shape[1]),
                            dtype=file["Subhalo/" + ikey].dtype)
                    else:
                        assert False

            # read subhalo data
            if ns > 0:
                data[skip_sub:skip_sub + ns] = file["Subhalo/" + ikey]

            skip_sub += ns

            file.close()

        # Load all variables with double precision
        if settings.double_precision:
            if not issubclass(data.dtype.type, numbers.Integral):
                data = data.astype(np.float64)
        else:
            warnings.warn('\n\nThe cython routines expect double precision '
                          + 'and will fail unless settings.double_precision '
                          + 'is True.\n\n')

        if settings.use_units:
            self.Sub[ikey] = self.get_paicos_quantity(
                data, ikey,
                field='subhalos')
            if not hasattr(self.Sub[ikey], 'unit'):
                del self.Sub[ikey]
                raise RuntimeError(f"{ikey} does not have units implemented!")
        else:
            self.Sub[ikey] = data

    def load_all_data(self):
        """
        Calling this method simply loads all the data in the catalog.
        """

        for ikey in self.Group._auto_list:
            self.load_group_data(ikey)

        for ikey in self.Sub._auto_list:
            self.load_sub_data(ikey)

    def save_new_catalog(self, basename, basedir=None, single_precision=False):
        """
        Save a new catalog containing only the currently loaded
        variables. Useful for reducing datasets to smaller sizes.
        """
        if basedir is None:
            writer = PaicosWriter(self, self.basedir, basename, 'w')
        else:
            writer = PaicosWriter(self, basedir, basename, 'w')

        for key in self.Group:
            Ngroups_Total = self.Group[key].shape[0]

            if single_precision:
                data = self.Group[key].astype(np.float32)
            else:
                data = self.Group[key]
            writer.write_data(key, data, group='Group')

        for key in self.Sub:
            Nsubgroups_Total = self.Sub[key].shape[0]

            if single_precision:
                data = self.Sub[key].astype(np.float32)
            else:
                data = self.Sub[key]
            writer.write_data(key, data, group='Subhalo')

        with h5py.File(writer.tmp_filename, 'r+') as f:
            f['Header'].attrs["Ngroups_ThisFile"] = Ngroups_Total
            f['Header'].attrs["Ngroups_Total"] = Ngroups_Total
            f['Header'].attrs["Nsubgroups_ThisFile"] = Nsubgroups_Total
            f['Header'].attrs["Nsubgroups_Total"] = Nsubgroups_Total
            f['Header'].attrs["NumFiles"] = 1

        writer.finalize()

        return writer



import cython
import numpy as np
cimport numpy as np
np.import_array()
from cython.parallel import prange
from libc.math cimport fmin, fmax
from libc.math cimport ceil
cimport cython
from libc.stdlib cimport malloc, free
from libc.string cimport memset

ctypedef np.int32_t int32_t
ctypedef np.int64_t int64_t


ctypedef unsigned long ulong
ctypedef double real_t

cdef extern int __builtin_clzl(unsigned long x) nogil

cdef int L = 21

def leading_zeros_cython(unsigned long x):
    return __builtin_clzl(x)

cdef api long leading_zeros_cython_for_numba(unsigned long x):
    return <long>(__builtin_clzl(x))

def get_len_of_common_prefix_cython(ulong key1, ulong key2):
    return __builtin_clzl(key1 ^ key2)

def get_leading_zeros(unsigned long[:] morton_keys):
    cdef int ii, N
    N = morton_keys.shape[0] - 1
    cdef int[:] my_leading_zeros = np.zeros(N, dtype=np.int32)
    for ii in range(N):
        key = morton_keys[ii] ^ morton_keys[ii+1]
        my_leading_zeros[ii] = __builtin_clzl(key)
    tmp = np.zeros(N, dtype=np.int32)
    tmp[:] = my_leading_zeros[:]
    return tmp

def set_leaf_bounding_volumes(ulong[:, :, :] tree_bounds,
                               real_t[:, :] points,
                               real_t[:] half_size,
                               real_t conversion_factor,
                               ulong L):
    # cdef ulong L = 21
    cdef Py_ssize_t ip
    cdef Py_ssize_t num_leafs = half_size.shape[0]
    cdef Py_ssize_t num_internal_nodes = num_leafs - 1

    cdef real_t f = conversion_factor
    cdef real_t cx, cy, cz, hs
    cdef ulong max_index = (<ulong>1 << L) - 1
    cdef ulong x_min, y_min, z_min, x_max, y_max, z_max

    with nogil:
        for ip in prange(num_leafs, schedule='static'):
            hs = f * half_size[ip]
            cx = points[ip, 0]
            cy = points[ip, 1]
            cz = points[ip, 2]

            x_min = <ulong> max(cx - hs, 0.0)
            y_min = <ulong> max(cy - hs, 0.0)
            z_min = <ulong> max(cz - hs, 0.0)

            x_max = <ulong> min(ceil(cx + hs), max_index)
            y_max = <ulong> min(ceil(cy + hs), max_index)
            z_max = <ulong> min(ceil(cz + hs), max_index)

            tree_bounds[ip + num_internal_nodes, 0, 0] = x_min
            tree_bounds[ip + num_internal_nodes, 1, 0] = y_min
            tree_bounds[ip + num_internal_nodes, 2, 0] = z_min
            tree_bounds[ip + num_internal_nodes, 0, 1] = x_max
            tree_bounds[ip + num_internal_nodes, 1, 1] = y_max
            tree_bounds[ip + num_internal_nodes, 2, 1] = z_max



def propagate_bounds_upwards(ulong[:, :, :] tree_bounds,
                             int64_t[:] tree_parents,
                             int64_t[:, :] tree_children):
    cdef int num_internal_nodes = tree_children.shape[0]
    cdef int num_leafs = num_internal_nodes + 1
    cdef int num_total_nodes = num_internal_nodes + num_leafs

    cdef int64_t *queue = <int64_t *> malloc(num_total_nodes * sizeof(int64_t))
    cdef int64_t *next_queue = <int64_t *> malloc(num_total_nodes * sizeof(int64_t))
    cdef int *seen = <int *> malloc(num_total_nodes * sizeof(int))
    cdef int *child_counter = <int *> malloc(num_total_nodes * sizeof(int))
    memset(seen, 0, num_total_nodes * sizeof(int))
    memset(child_counter, 0, num_total_nodes * sizeof(int))

    cdef int i, node_id, parent_id
    cdef int queue_len = 0
    cdef int next_len = 0

    # Queue initialization: find unique parents of leaves
    for i in range(num_leafs):
        node_id = num_internal_nodes + i
        parent_id = tree_parents[node_id]
        if parent_id >= 0:
            child_counter[parent_id] += 1

    for i in range(num_internal_nodes):
        if child_counter[i] == 2:
            seen[i] = 1
            queue[queue_len] = i
            queue_len += 1

    cdef ulong[:, :, :] bounds = tree_bounds
    cdef int64_t[:] parents = tree_parents
    cdef int64_t[:, :] children = tree_children
    cdef int childA, childB, dim
    cdef ulong a_min, b_min, a_max, b_max

    while queue_len > 0:
        next_len = 0
        for i in range(queue_len):
            node_id = queue[i]
            parent_id = parents[node_id]
            if parent_id >= 0:
                child_counter[parent_id] += 1
                if child_counter[parent_id] == 2 and seen[parent_id] == 0:
                    seen[parent_id] = 1
                    next_queue[next_len] = parent_id
                    next_len += 1

            childA = children[node_id, 0]
            childB = children[node_id, 1]

            for dim in range(3):
                a_min = bounds[childA, dim, 0]
                b_min = bounds[childB, dim, 0]
                a_max = bounds[childA, dim, 1]
                b_max = bounds[childB, dim, 1]
                bounds[node_id, dim, 0] = min(a_min, b_min)
                bounds[node_id, dim, 1] = max(a_max, b_max)

        queue_len = next_len
        tmp = queue
        queue = next_queue
        next_queue = tmp

    free(queue)
    free(next_queue)
    free(seen)
    free(child_counter)

cdef inline int count_leading_zeros(unsigned long x) noexcept nogil:
    return __builtin_clzl(x)

cdef inline ulong part1by2_64(ulong n) noexcept nogil:
    n &= <ulong>0x1fffff
    n = (n | (n << 32)) & <ulong>0x1f00000000ffff
    n = (n | (n << 16)) & <ulong>0x1f0000ff0000ff
    n = (n | (n << 8)) & <ulong>0x100f00f00f00f00f
    n = (n | (n << 4)) & <ulong>0x10c30c30c30c30c3
    n = (n | (n << 2)) & <ulong>0x1249249249249249
    return n

cdef inline ulong unpart1by2_64(ulong n) noexcept nogil:
    n &= <ulong>0x1249249249249249
    n = (n ^ (n >> 2)) & <ulong>0x10c30c30c30c30c3
    n = (n ^ (n >> 4)) & <ulong>0x100f00f00f00f00f
    n = (n ^ (n >> 8)) & <ulong>0x1f0000ff0000ff
    n = (n ^ (n >> 16)) & <ulong>0x1f00000000ffff
    n = (n ^ (n >> 32)) & <ulong>0x1fffff
    return n

cdef inline ulong encode64(ulong x, ulong y, ulong z) noexcept nogil:
    return 4 * part1by2_64(x) + 2 * part1by2_64(y) + part1by2_64(z)

def encode64_cython(ulong x, ulong y, ulong z):
    return 4 * part1by2_64(x) + 2 * part1by2_64(y) + part1by2_64(z)

cdef inline void decode64(ulong key, ulong[:] out) noexcept nogil:
    out[0] = unpart1by2_64(key >> 2)
    out[1] = unpart1by2_64(key >> 1)
    out[2] = unpart1by2_64(key)

def get_morton_keys64(ulong[:, :] pos):
    cdef Py_ssize_t N = pos.shape[0]
    cdef ulong[:] morton_keys = np.empty(N, dtype=np.uint64)
    cdef Py_ssize_t i
    for i in range(N):
        morton_keys[i] = encode64(<ulong>pos[i,0], <ulong>pos[i,1], <ulong>pos[i,2])

    tmp = np.empty(N, dtype=np.uint64)
    tmp[:] = morton_keys[:]
    return tmp

cdef inline int delta(ulong[:] sortedMortonCodes, int i, int j) noexcept nogil:
    cdef ulong codeA = sortedMortonCodes[i]
    cdef ulong codeB = sortedMortonCodes[j]

    if codeA == codeB:
        # Fallback: compare indices instead of codes
        return 64 + count_leading_zeros(<ulong>(i ^ j))
    else:
        return count_leading_zeros(codeA ^ codeB)

cdef inline int findSplit(ulong[:] sortedMortonCodes, int first, int last) noexcept nogil:
    cdef int commonPrefix = delta(sortedMortonCodes, first, last)
    cdef int split = first
    cdef int step = last - first
    cdef int newSplit, splitPrefix

    while step > 1:
        step = (step + 1) >> 1
        newSplit = split + step
        if newSplit < last:
            splitPrefix = delta(sortedMortonCodes, first, newSplit)
            if splitPrefix > commonPrefix:
                split = newSplit

    return split

cdef inline void determineRange(ulong[:] sortedMortonCodes, int n_codes, int idx, int[:] first_last) noexcept nogil:
    cdef int d = 1, minPrefixLength = -1
    cdef int firstIndex = idx, max_last = n_codes
    cdef int lz_p1, lz_m1, lz_first_second, searchRange, secondIndex, newJdx
    cdef int first, last

    if (firstIndex > 0):
        # Count leading zeros
        lz_p1 = delta(sortedMortonCodes, firstIndex, firstIndex + 1)
        lz_m1 = delta(sortedMortonCodes, firstIndex, firstIndex - 1)

        if lz_p1 > lz_m1:
            d = 1
            minPrefixLength = lz_m1
        else:
            d = -1
            minPrefixLength = lz_p1

    searchRange = 2
    secondIndex = firstIndex + searchRange * d

    while (0 <= secondIndex and secondIndex < max_last):
        lz_first_second = delta(sortedMortonCodes, firstIndex, secondIndex)
        if lz_first_second > minPrefixLength:
            searchRange *= 2
            secondIndex = firstIndex + searchRange * d
            if (0 <= secondIndex and secondIndex < max_last):
                lz_first_second = delta(sortedMortonCodes, firstIndex, secondIndex)
            else:
                break
        else:
            break

    # start binary search with known searchRange
    secondIndex = firstIndex
    while searchRange > 1:
        searchRange = (searchRange + 1) // 2
        newJdx = secondIndex + searchRange * d
        if (0 <= newJdx and newJdx < max_last):
            lz_f_Jdx = delta(sortedMortonCodes, firstIndex, newJdx)
            if lz_f_Jdx > minPrefixLength:
                secondIndex = newJdx

    first = min(secondIndex, firstIndex)
    last = max(secondIndex, firstIndex)
    first_last[0] = first
    first_last[1] = last

def generateHierarchy(ulong[:] sortedMortonCodes, long[:, :] tree_children, long[:] tree_parents):
    cdef int n_codes = sortedMortonCodes.shape[0]
    cdef int num_internal_nodes = n_codes - 1
    cdef int idx, first, last, split
    cdef int childA, childB

    cdef int[:] first_last = np.empty(2, dtype=np.int32)
    # cdef int[2] first_last

    for idx in range(num_internal_nodes):
        determineRange(sortedMortonCodes, n_codes, idx, first_last)
        first = first_last[0]
        last = first_last[1]
        split = findSplit(sortedMortonCodes, first, last)

        # Select childA
        if split == first:
            # Is a leaf
            childA = split + num_internal_nodes
        else:
            # Is in an internal node
            childA = split

        # Select childB
        if split + 1 == last:
            childB = split + 1 + num_internal_nodes
        else:
            childB = split + 1

        # Set children and parent
        tree_children[idx, 0] = childA
        tree_children[idx, 1] = childB
        tree_parents[childA] = idx
        tree_parents[childB] = idx



from cython.parallel import prange, parallel
from libc.math cimport fmin, fmax, sqrt
from libc.stdlib cimport malloc, free
cimport openmp
import numpy as np
cimport numpy as np

ctypedef double real_t

cdef inline int rotate_point_around_center(real_t[:] point,
                                            real_t[:] tmp_point,
                                            real_t[:] center,
                                            real_t[:, :] rotation_matrix) noexcept nogil:
    cdef int ii, jj

    # Subtract center and zero out point
    for ii in range(3):
        tmp_point[ii] = point[ii] - center[ii]
        point[ii] = 0.0

    # Matrix-vector multiplication
    for ii in range(3):
        for jj in range(3):
            point[ii] += rotation_matrix[ii, jj] * tmp_point[jj]

    # Add center back
    for ii in range(3):
        point[ii] += center[ii]

    return 0


cdef inline real_t distance(real_t[:] p1, real_t[:] p2) noexcept nogil:
    cdef real_t dx = p1[0] - p2[0]
    cdef real_t dy = p1[1] - p2[1]
    cdef real_t dz = p1[2] - p2[2]
    return sqrt(dx*dx + dy*dy + dz*dz)

cdef inline real_t distance_to_box(real_t[:] point, unsigned long[:, :] box) noexcept nogil:
    cdef real_t d
    cdef real_t sq_dist = 0.0
    cdef int i
    for i in range(3):
        if point[i] < box[i, 0]:
            d = box[i, 0] - point[i]
            sq_dist += d * d
        elif point[i] > box[i, 1]:
            d = point[i] - box[i, 1]
            sq_dist += d * d
    return sqrt(sq_dist)

cdef inline int nearest_neighbor(real_t[:, :] points, long[:] tree_parents,
                                   long[:, :] tree_children, unsigned long[:, :, :] tree_bounds,
                                   real_t[:] query_point, int num_internal_nodes, real_t[:] min_dist_final) noexcept nogil:
    cdef int queue[128]
    cdef int queue_index = 0
    queue[queue_index] = 0
    cdef int L = 21
    cdef real_t min_dist = (2 ** L - 1.0)
    cdef int min_index = -1
    cdef int node_id, childA, childB, data_id
    cdef bint is_leafA, is_leafB
    cdef real_t dist, distA, distB
    cdef bint traverseA, traverseB

    while queue_index >= 0:
        node_id = queue[queue_index]
        childA = tree_children[node_id, 0]
        childB = tree_children[node_id, 1]

        is_leafA = childA >= num_internal_nodes
        is_leafB = childB >= num_internal_nodes

        if is_leafA:
            data_id = childA - num_internal_nodes
            dist = distance(points[data_id], query_point)
            if dist < min_dist:
                min_dist = dist
                min_index = data_id

        if is_leafB:
            data_id = childB - num_internal_nodes
            dist = distance(points[data_id], query_point)
            if dist < min_dist:
                min_dist = dist
                min_index = data_id

        distA = distance_to_box(query_point, tree_bounds[childA])
        distB = distance_to_box(query_point, tree_bounds[childB])

        traverseA = (distA <= min_dist) and not is_leafA
        traverseB = (distB <= min_dist) and not is_leafB

        if not traverseA and not traverseB:
            queue_index -= 1
        else:
            if traverseA:
                queue[queue_index] = childA
            else:
                queue[queue_index] = childB
            if traverseA and traverseB:
                queue_index += 1
                queue[queue_index] = childB
    min_dist_final[0] = min_dist
    return min_index

cdef inline bint is_point_in_box(real_t[:] point, unsigned long[:, :] box) noexcept nogil:
    cdef int ii
    for ii in range(3):
        if point[ii] < box[ii, 0] or point[ii] > box[ii, 1]:
            return False
    return True

cdef inline int nearest_neighbor_voronoi(real_t[:, :] points, long[:] tree_parents,
                                   long[:, :] tree_children, unsigned long[:, :, :] tree_bounds,
                                   real_t[:] query_point, int num_internal_nodes, real_t[:] min_dist_final) noexcept nogil:
    cdef int queue[128]
    cdef int queue_index = 0
    queue[queue_index] = 0
    cdef int L = 21
    cdef real_t min_dist = (2 ** L - 1.0)
    cdef int min_index = -1
    cdef int node_id, childA, childB, data_id
    cdef bint is_leafA, is_leafB, point_in_A, point_in_B
    cdef real_t dist, distA, distB
    cdef bint traverseA, traverseB

    while queue_index >= 0:
        node_id = queue[queue_index]
        childA = tree_children[node_id, 0]
        childB = tree_children[node_id, 1]

        is_leafA = childA >= num_internal_nodes
        is_leafB = childB >= num_internal_nodes

        point_in_A = is_point_in_box(query_point, tree_bounds[childA])
        point_in_B = is_point_in_box(query_point, tree_bounds[childB])

        if point_in_A and is_leafA:
            data_id = childA - num_internal_nodes
            dist = distance(points[data_id], query_point)
            if dist < min_dist:
                min_dist = dist
                min_index = data_id

        if point_in_B and is_leafB:
            data_id = childB - num_internal_nodes
            dist = distance(points[data_id], query_point)
            if dist < min_dist:
                min_dist = dist
                min_index = data_id

        # Whether to traverse
        traverseA = point_in_A and not is_leafA
        traverseB = point_in_B and not is_leafB

        if not traverseA and not traverseB:
            queue_index -= 1
        else:
            if traverseA:
                queue[queue_index] = childA
            else:
                queue[queue_index] = childB
            if traverseA and traverseB:
                queue_index += 1
                queue[queue_index] = childB
    min_dist_final[0] = min_dist
    return min_index

# from cython.parallel cimport prange, parallel
# from libc.math cimport fmin, fmax

# # Assume real_t is already defined, e.g.
# # ctypedef double real_t

# from cython.parallel cimport prange, parallel
# from libc.math cimport sqrt
# cimport cython

# # Assume: real_t is already defined (e.g., ctypedef double real_t)
# # Also assume: nearest_neighbor() and rotate_point_around_center() are declared elsewhere and cimported

def trace_rays_cpu(real_t[:, :] points,
                   long[:] tree_parents,
                   long[:, :] tree_children,
                   unsigned long[:, :, :] tree_bounds,
                   real_t[:] variable,
                   real_t[:] hsml,
                   real_t[:] widths,
                   real_t[:] center,
                   real_t tree_scale_factor,
                   real_t[:] tree_offsets,
                   real_t[:, :] image,
                   real_t[:, :] rotation_matrix,
                   real_t tol,
                   int numthreads):

    cdef int nx = image.shape[0]
    cdef int ny = image.shape[1]
    cdef real_t dx = widths[0] / nx
    cdef real_t dy = widths[1] / ny

    cdef int ix, iy, i
    cdef real_t z, dz, result
    cdef real_t min_dist
    cdef int min_index

    cdef int num_internal_nodes = tree_children.shape[0]

    cdef real_t[:] query_point = np.zeros((3), dtype=np.float64)
    cdef real_t[:] tmp_point = np.zeros((3), dtype=np.float64)
    cdef real_t[:] min_dist_final = np.zeros((1), dtype=np.float64)

    # with nogil, parallel(num_threads=numthreads):
    # for ix in range(nx, schedule='static'):
    for ix in range(nx):
        for iy in range(ny):
            result = 0.0
            z = 0.0

            while z < widths[2]:
                # Compute query point in world coordinates
                query_point[0] = (center[0] - widths[0] / 2.0) + (ix + 0.5) * dx
                query_point[1] = (center[1] - widths[1] / 2.0) + (iy + 0.5) * dy
                query_point[2] = (center[2] - widths[2] / 2.0) + z

                # Rotate point around center (in-place)
                rotate_point_around_center(query_point, tmp_point, center, rotation_matrix)

                # Transform to tree coordinates
                for i in range(3):
                    query_point[i] = (query_point[i] - tree_offsets[i]) * tree_scale_factor

                # Nearest neighbor
                min_index = nearest_neighbor(points, tree_parents, tree_children,
                                            tree_bounds, query_point, num_internal_nodes, min_dist_final)

                # min_dist = distance(points[min_index], query_point)
                min_dist = min_dist_final[0]

                # Adaptive integration step (convert to Arepo units)
                dz = tol * min_dist / tree_scale_factor
                result = result + dz * variable[min_index]
                z = z + dz

            # Overshoot correction
            result -= (z - widths[2]) * variable[min_index]
            image[ix, iy] = result


def trace_rays_cpu_voronoi(real_t[:, :] points,
                   long[:] tree_parents,
                   long[:, :] tree_children,
                   unsigned long[:, :, :] tree_bounds,
                   real_t[:] variable,
                   real_t[:] hsml,
                   real_t[:] widths,
                   real_t[:] center,
                   real_t tree_scale_factor,
                   real_t[:] tree_offsets,
                   real_t[:, :] image,
                   real_t[:, :] rotation_matrix,
                   real_t tol,
                   int numthreads):

    cdef int nx = image.shape[0]
    cdef int ny = image.shape[1]
    cdef real_t dx = widths[0] / nx
    cdef real_t dy = widths[1] / ny

    cdef int ix, iy, i
    cdef real_t z, dz, result
    cdef real_t min_dist
    cdef int min_index

    cdef int num_internal_nodes = tree_children.shape[0]

    cdef real_t[:] query_point = np.zeros((3), dtype=np.float64)
    cdef real_t[:] tmp_point = np.zeros((3), dtype=np.float64)
    cdef real_t[:] min_dist_final = np.zeros((1), dtype=np.float64)

    # with nogil, parallel(num_threads=numthreads):
    # for ix in range(nx, schedule='static'):
    for ix in range(nx):
        for iy in range(ny):
            result = 0.0
            z = 0.0

            while z < widths[2]:
                # Compute query point in world coordinates
                query_point[0] = (center[0] - widths[0] / 2.0) + (ix + 0.5) * dx
                query_point[1] = (center[1] - widths[1] / 2.0) + (iy + 0.5) * dy
                query_point[2] = (center[2] - widths[2] / 2.0) + z

                # Rotate point around center (in-place)
                rotate_point_around_center(query_point, tmp_point, center, rotation_matrix)

                # Transform to tree coordinates
                for i in range(3):
                    query_point[i] = (query_point[i] - tree_offsets[i]) * tree_scale_factor

                # Nearest neighbor
                min_index = nearest_neighbor_voronoi(points, tree_parents, tree_children,
                                            tree_bounds, query_point, num_internal_nodes, min_dist_final)
                if min_index == -1:
                    min_index = nearest_neighbor(points, tree_parents, tree_children,
                                            tree_bounds, query_point, num_internal_nodes, min_dist_final)


                # min_dist = distance(points[min_index], query_point)
                min_dist = min_dist_final[0]

                # Adaptive integration step (convert to Arepo units)
                dz = tol * hsml[min_index]
                result = result + dz * variable[min_index]
                z = z + dz

            # Overshoot correction
            result -= (z - widths[2]) * variable[min_index]
            image[ix, iy] = result



import numpy as np
from ..writers.arepo_image import ImageWriter
from ..readers.paicos_readers import ImageReader
import os


def find_angle_subdivisions(angle, max_angle):
    kk = 1
    angle = abs(angle)
    new_angle = angle / kk
    while new_angle > max_angle:
        kk += 1
        new_angle = angle / kk

    return kk


def find_zoom_sub_divisions(zoom, zoom_max):
    kk = 1
    if zoom < 1:
        zoom = 1 / zoom
    new_zoom = zoom**(1 / kk)
    while new_zoom > zoom_max:
        kk += 1
        new_zoom = zoom**(1 / kk)

    return kk


class Actions:
    """
    The purpose of this class is to simplify the steps
    that are usually involved in creating animations
    and/or interactive widgets.

    This is very much work in progress and subject
    to change. Send Thomas an email if you would like
    a working example.
    """
    def __init__(self, snapnum=None, image_creator=None, dry_run=True):

        self.dry_run = dry_run

        self.first_call = True
        if snapnum is None:
            if image_creator is not None:
                self.image_creator = image_creator
            else:
                msg = ("self.image_creator not set, you'll "
                       + "nedd to do this manually or read "
                       + "a log file using Actions.read_log")
                print(msg)
        else:
            self.make_image_creator(snapnum)
        self.first_call = False

        self.logging = False
        self.zoom_max = 10000000.
        self.max_angle = 360.

    def make_image_creator(self, snapnum):
        """

        TODO: Add example implementation in the docstring of
        of this method.
        """
        err_msg = ("You need to implement this yourself.\n"
                   + "Please see the docstring of this method")
        raise RuntimeError(err_msg)

    def mylogger(self, string, line_ending='\n', mode='a', command=True):
        outfolder = self.outfolder
        basename = self.basename
        with open(f'{outfolder}/{basename}.log', mode) as f:
            f.write(string + line_ending)

        if self.make_waypoints and command:
            self.frame_num += 1
            waypoint_folder = f'{self.outfolder}/{self.basename}_waypoints'
            self._make_waypoint(waypoint_folder, self.basename, self.frame_num)

    def create_log(self, outfolder='.', basename='image'):

        if outfolder[-1] != '/':
            outfolder += '/'
        if not os.path.exists(outfolder):
            os.makedirs(outfolder)

        self.make_waypoints = True
        self.frame_num = 0
        self.outfolder = outfolder
        self.basename = basename
        self.mylogger('# Paicos image log file', mode='w', command=False)
        self.mylogger(outfolder, command=False)
        self.mylogger(basename, command=False)

        image_file = ImageWriter(self.image_creator, basedir=outfolder,
                                 basename=f'{basename}_logfile_start')
        image_file.finalize()
        self.mylogger(image_file.filename, command=False)
        self.logging = True

        if self.make_waypoints:
            waypoint_folder = f'{self.outfolder}{self.basename}_waypoints/'
            self._make_waypoint(waypoint_folder, self.basename, 0)

    def read_log(self, filename=None, outfolder='.', basename='image'):
        self.make_waypoints = False
        if filename is None:
            filename = f'{outfolder}/{basename}.log'
        with open(filename, 'r') as f:
            lines = f.readlines()
        self.outfolder = outfolder = lines[1].strip()
        self.basename = basename = lines[2].strip()
        self.logfile_start = lines[3].strip()
        self.lines = lines
        self.commands = self.lines[4:]
        self.frame_num = 0
        self.line_index = 4
        self.num_steps = len(self.lines) - self.line_index
        self.image_creator = ImageReader(self.logfile_start)
        self.first_call = False
        self.make_image_creator(self.image_creator.snapnum)

        waypoint_folder = f'{self.outfolder}/{self.basename}_waypoints/'
        if os.path.exists(waypoint_folder):
            import glob
            files = glob.glob(f'{waypoint_folder}*.hdf5')
            framenums = [int(file.split('_')[-2]) for file in files]
            snapnums = [int(file[-8:-5]) for file in files]
            index = np.argsort(framenums)
            files = [files[ii] for ii in index]
            self.waypoints = {}
            framenums = np.array(framenums)[index]
            snapnums = np.array(snapnums)[index]
            for ii, frame_num in enumerate(framenums):
                self.waypoints[frame_num] = files[ii]

    def resume_from_waypoint(self, frame_num):
        self.frame_num = frame_num
        self.line_index = self.frame_num + 4

        self.image_creator = ImageReader(self.waypoints[frame_num])
        self.first_call = False
        self.make_image_creator(self.image_creator.snapnum)

    def resume_from_image(self, filename, frame_num):
        raise RuntimeError('not implemented, try resume_from_waypoint instead')

    @property
    def next_action(self):
        return self.lines[self.line_index].strip()

    @property
    def previous_action(self):
        if self.line_index == 4:
            return None
        else:
            return self.lines[self.line_index - 1].strip()

    def step(self, verbose=False):
        line = self.next_action
        parts = line.split(',')
        command = parts[0]

        im_creator = self.image_creator

        if verbose:
            print(line)
        if command == 'center_on_sub' or command == 'center_on_group':
            # cx, cy, cz = float(parts[2]), float(parts[3]), float(parts[4])
            # new_center = np.array([cx, cy, cz]) * im_creator.center.uq
            raise RuntimeError("not implemented")
            # self.reset_center(new_center)
        elif command == 'zoom':
            zoom_fac = float(parts[1])
            if zoom_fac > self.zoom_max or (1 / zoom_fac) > self.zoom_max:
                kk = find_zoom_sub_divisions(zoom_fac, self.zoom_max)
                new_zoom_fac = zoom_fac**(1 / kk)
                for _ in range(kk):
                    self.zoom(new_zoom_fac)
            else:
                self.zoom(zoom_fac)
        elif command == 'width':
            self.width(float(parts[1]))
        elif command == 'height':
            self.height(float(parts[1]))
        elif command == 'depth':
            self.depth(float(parts[1]))
        elif command == 'rotate_around_perp_vector1':
            angle = float(parts[1])
            if abs(angle) > self.max_angle:
                kk = find_angle_subdivisions(angle, self.max_angle)
                for _ in range(kk):
                    self.rotate_around_perp_vector1(angle / kk)
            else:
                self.rotate_around_perp_vector1(angle)

        elif command == 'rotate_around_perp_vector2':
            angle = float(parts[1])
            if abs(angle) > self.max_angle:
                kk = find_angle_subdivisions(angle, self.max_angle)
                for _ in range(kk):
                    self.rotate_around_perp_vector2(angle / kk)
            else:
                self.rotate_around_perp_vector2(angle)
        elif command == 'rotate_around_normal_vector':
            angle = float(parts[1])
            if abs(angle) > self.max_angle:
                kk = find_angle_subdivisions(angle, self.max_angle)
                for _ in range(kk):
                    self.rotate_around_normal_vector(angle / kk)
            else:
                self.rotate_around_normal_vector(angle)
        elif command == 'half_resolution':
            self.half_resolution()
        elif command == 'double_resolution':
            self.double_resolution()
        elif command == 'move_center':
            diff_x, diff_y, diff_z = float(
                parts[1]), float(parts[2]), float(parts[3])
            diff = np.array([diff_x, diff_y, diff_z]) * im_creator.center.uq
            self.move_center(diff)
        elif command == 'move_center_sim_coordinates':
            diff_x, diff_y, diff_z = float(
                parts[1]), float(parts[2]), float(parts[3])
            diff = np.array([diff_x, diff_y, diff_z]) * im_creator.center.uq
            self.move_center_sim_coordinates(diff)
        elif command == 'reset_center':
            self.reset_center()
        elif command == 'change_snapnum':
            self.change_snapnum(int(parts[1]))
        else:
            raise RuntimeError(f'unknown command in log file!\n{line}\n{command}')

        if not self.logging:
            self.frame_num += 1
        self.line_index += 1

    def _make_waypoint(self, waypoint_folder, basename, frame_num):

        image_file = ImageWriter(self.image_creator, basedir=waypoint_folder,
                                 basename=f'waypoint_{basename}_frame_{frame_num}')
        image_file.finalize()

    def expand_log(self, zoom_max=1.02, max_angle=1, verbose=False, outfolder=None, basename=None):
        self.zoom_max = zoom_max
        self.max_angle = max_angle

        self.read_log(outfolder=self.outfolder, basename=self.basename)

        if outfolder is not None:
            self.outfolder = outfolder
        if basename is not None:
            self.basename = basename

        self.basename = self.basename + '_expanded'

        self.create_log(outfolder=self.outfolder, basename=self.basename)

        for ii in range(4, len(self.lines)):
            self.step(verbose=verbose)

    def change_snapnum(self, snapnum):
        self.make_image_creator(snapnum)

        if self.logging:
            self.mylogger(f'change_snapnum,{snapnum}')

    def zoom(self, factor):
        im_creator = self.image_creator
        im_creator.zoom(factor)
        if self.logging:
            self.mylogger(f'zoom,{factor}')

    def reset_center(self):
        self.image_creator.reset_center()
        if self.logging:
            self.mylogger("reset_center")

    # def center_on_sub(self, sub_id):
    #     im_creator = self.image_creator
    #     new_center = im_creator.snap.Cat.Sub['SubhaloPos'][sub_id].T
    #     im_creator._diff_center += new_center - im_creator._center
    #     im_creator.center = new_center.copy

    # def center_on_group(self, group_id):
    #     im_creator = self.image_creator
    #     new_center = im_creator.snap.Cat.Group['GroupPos'][group_id].T
    #     im_creator._diff_center += new_center - im_creator._center
    #     im_creator.center = new_center.copy

    def width(self, arg):
        im_creator = self.image_creator
        im_creator.width = arg * im_creator.width.uq
        if self.logging:
            self.mylogger(f'width,{arg}')

    def height(self, arg):
        im_creator = self.image_creator
        im_creator.height = arg * im_creator.height.uq
        if self.logging:
            self.mylogger(f'height,{arg}')

    def depth(self, arg):
        im_creator = self.image_creator
        im_creator.depth = arg * im_creator.depth.uq
        if self.logging:
            self.mylogger(f'depth,{arg}')

    def half_resolution(self):
        im_creator = self.image_creator
        im_creator.half_resolution
        if self.logging:
            self.mylogger('half_resolution')

    def double_resolution(self):
        im_creator = self.image_creator
        im_creator.double_resolution
        if self.logging:
            self.mylogger('double_resolution')

    def rotate_around_perp_vector1(self, arg):
        im_creator = self.image_creator
        im_creator.orientation.rotate_around_perp_vector1(degrees=arg)
        if self.logging:
            self.mylogger(f'rotate_around_perp_vector1,{arg}')

    def rotate_around_perp_vector2(self, arg):
        im_creator = self.image_creator
        im_creator.orientation.rotate_around_perp_vector2(degrees=arg)
        if self.logging:
            self.mylogger(f'rotate_around_perp_vector2,{arg}')

    def rotate_around_normal_vector(self, arg):
        im_creator = self.image_creator
        im_creator.orientation.rotate_around_normal_vector(degrees=arg)
        if self.logging:
            self.mylogger(f'rotate_around_normal_vector,{arg}')

    def move_center(self, diff):
        im_creator = self.image_creator
        im_creator.move_center_along_perp_vector1(diff[0])
        im_creator.move_center_along_perp_vector2(diff[1])
        im_creator.move_center_along_normal_vector(diff[2])
        if self.logging:
            self.mylogger(f"move_center,{diff[0].value},{diff[1].value},{diff[2].value}")

    def move_center_sim_coordinates(self, diff):
        im_creator = self.image_creator
        new_center = diff + im_creator.center
        im_creator.center = new_center
        im_creator._diff_center += diff
        if self.logging:
            self.mylogger(f"move_center_sim_coordinates,{diff[0].value},{diff[1].value},{diff[2].value}")
